import networkx as nx
import matplotlib.pyplot as plt

class graphDatabase:

    def __init__(self, iFile):
        self.graphs = []
        current_graph = {'vertices': [], 'edges': []}

        with open(iFile, 'r') as file:
            for line in file:
                if line.startswith('t #'):
                    if current_graph['vertices'] or current_graph['edges']:
                        self.graphs.append(current_graph)
                        current_graph = {'vertices': [], 'edges': []}
                elif line.startswith('v'):
                    _, v_id, label = line.split()
                    current_graph['vertices'].append((int(v_id), int(label)))
                elif line.startswith('e'):
                    _, v1, v2, label = line.split()
                    current_graph['edges'].append((int(v1), int(v2), int(label)))

        if current_graph['vertices'] or current_graph['edges']:
            self.graphs.append(current_graph)

    def printIndividualGraphStats(self):
        for i, graph in enumerate(self.graphs):
            print(f"Graph {i}:")
            num_vertices = len(graph['vertices'])
            num_edges = len(graph['edges'])
            vertex_labels = set(label for _, label in graph['vertices'])
            edge_labels = set(label for _, _, label in graph['edges'])

            print(f"  Number of vertices: {num_vertices}")
            print(f"  Number of edges: {num_edges}")
            print(f"  Unique vertex labels: {vertex_labels}")
            print(f"  Unique edge labels: {edge_labels}")

    def printGraphDatabaseStatistics(self):
        total_nodes = 0
        total_edges = 0
        vertex_labels = set()
        edge_labels = set()
        
        self.nodes_per_graph = [len(graph['vertices']) for graph in self.graphs]
        self.edges_per_graph = [len(graph['edges']) for graph in self.graphs]
        
        for graph in self.graphs:
            total_nodes += len(graph['vertices'])
            total_edges += len(graph['edges'])
            
            for vertex in graph['vertices']:
                vertex_labels.add(vertex[1])  
                
            for edge in graph['edges']:
                edge_labels.add(edge[2])  

        average_nodes = sum(self.nodes_per_graph) / len(self.graphs) if self.graphs else 0
        average_edges = sum(self.edges_per_graph) / len(self.graphs) if self.graphs else 0
        max_nodes = max(self.nodes_per_graph) if self.graphs else 0
        min_nodes = min(self.nodes_per_graph) if self.graphs else 0
        max_edges = max(self.edges_per_graph) if self.graphs else 0
        min_edges = min(self.edges_per_graph) if self.graphs else 0
        total_unique_vertex_labels = len(vertex_labels)
        total_unique_edge_labels = len(edge_labels)

        print(f'average_nodes: {average_nodes}')
        print(f'average_edges: {average_edges}')
        print(f'max_nodes: {max_nodes}')
        print(f'min_nodes: {min_nodes}')
        print(f'max_edges: {max_edges}')
        print(f'min_edges: {min_edges}')
        print(f'total_unique_vertex_labels: {total_unique_vertex_labels}')
        print(f'total_unique_edge_labels: {total_unique_edge_labels}')

    def plotNodeDistribution(self):
        
        plt.figure(figsize=(6, 4))
        plt.hist(self.nodes_per_graph, bins=max(20, len(set(self.nodes_per_graph))), edgecolor='black')
        plt.title('Distribution of Nodes per Graph')
        plt.xlabel('Number of Nodes')
        plt.ylabel('Frequency')
        plt.tight_layout()
        plt.show()

    def plotEdgeDistribution(self):
        
        plt.figure(figsize=(6, 4))
        plt.hist(self.edges_per_graph, bins=max(20, len(set(self.edges_per_graph))), edgecolor='black')
        plt.title('Distribution of Edges per Graph')
        plt.xlabel('Number of Edges')
        plt.ylabel('Frequency')
        plt.tight_layout()
        plt.show()
        

if __name__ == '__main__':
    file_path = 'Chemical_340.txt'
    obj = graphDatabase(file_path)
    obj.printGraphDatabaseStatistics()
    obj.printIndividualGraphStats()
    obj.plotNodeDistribution()
    obj.plotEdgeDistribution()


# DF2DB in this code dataframe is converting databases into sparse or dense transactional, temporal, Utility.
#
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.DF2DB import DF2DB as db
#
#             obj = db.DF2DB(idf, ">=", 16, "sparse/dense")
#
#             obj.getTransactional("outputFileName") # To create transactional database
#
#             obj.getTemporal("outputFileName") # To create temporal database
#
#             obj.getUtility("outputFileName") # To create utility database
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from PAMI.extras.DF2DB.denseDF2DB import *
from PAMI.extras.DF2DB.sparseDF2DB import *
import sys

class DF2DB:
    """
    :Description:  This class will create database for given DataFrame based on Threshold values and conditions are defined in the class.
                   Converts Dataframe into sparse or dense dataframes.

    :Attributes:

        :param inputDF: DataFrame :
             It is sparse or dense DataFrame
        :param thresholdValue: int or float :
             It is threshold value of all item
        :param condition: str :
             It is condition of all item
        :param DFtype: str :
             It is DataFrame type. It should be sparse or dense. Default DF is sparse.


    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.DF2DB import DF2DB as db

            obj = db.DF2DB(idf, ">=", 16, "sparse/dense")

            obj.getTransactional("outputFileName") # To create transactional database

            obj.getTemporal("outputFileName") # To create temporal database

            obj.getUtility("outputFileName") # To create utility database
    """


    def __init__(self, inputDF, thresholdValue, condition, DFtype='sparse') -> None:
        self.inputDF = inputDF
        self.thresholdValue = thresholdValue
        self.condition = condition
        self.DFtype = DFtype.lower()
        if DFtype == 'sparse':
            self.DF2DB = sparseDF2DB(self.inputDF, self.condition, self.thresholdValue)
        elif DFtype == 'dense':
            self.DF2DB = denseDF2DB(self.inputDF, self.condition, self.thresholdValue)
        else:
            raise Exception('DF type should be sparse or dense')

    def getTransactionalDatabase(self, outputFile) -> str:
        """
        create transactional database and return outputFileName
        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        :rtype: str
        """
        self.DF2DB.createTransactional(outputFile)
        return self.DF2DB.getFileName()

    def getTemporalDatabase(self, outputFile) -> str:
        """
        create temporal database and return outputFile name
        :param outputFile: file name or path to store database
        :type outputFile: str
        :return: outputFile name
        :rtype: str
        """
        self.DF2DB.createTemporal(outputFile)
        return self.DF2DB.getFileName()

    def getUtilityDatabase(self, outputFile) -> str:
        """
        create utility database and return outputFile name
        :param outputFile:  file name or path to store database
        :type outputFile: str
        :return: outputFile name
        :rtype: str
        """
        self.DF2DB.createUtility(outputFile)
        return self.DF2DB.getFileName()


if __name__ == '__main__':
    obj = DF2DB(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])
    obj.getTransactionalDatabase(sys.argv[5])


# csvParquet is a code used to convert temporal and utility types into sparse and dense format.
#
# **Importing this algorithm into a python program**
#
#             from PAMI.extras.csvParquet import csvParquet or Parquet2CSV as cp
#
#             obj = cp.CSV2Parquet(sampleDB.csv, output.parquet, sep)
#
#             obj = cp.Parquet2CSV(input.parquet, sampleDB.csv, sep)
#
#             obj.convert()
#
#             obj.printStats()
#

__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# import pandas as pd
# import numpy as np

# class CSV2Parquet():
#     """

#     :Description:   csvParquet is a code used to convert temporal and utility types into sparse and dense format

#     :param  iFile: str :
#                    Name of the Input file
#     :param  sep: str :
#                     This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

#     **Importing this algorithm into a python program**
#     --------------------------------------------------------
#     .. code-block:: python

#             from PAMI.extras.csvParquet import csvParquet as cp

#             obj = cp.csvParquet(File, "\t", " ", " " )

#             obj.save()

#             obj.csvParquet("FileName") # To generate file in form of sparse or dense

#             obj.parquetFormat("FileName") # To generate file in form of sparse or dense
#     """

#     def __init__(self, iFile: str, sep: str='\t'):
#         self._iFile = iFile
#         self._sep = sep

# def CSV2Parquet(csv_file: str, sep: str, inputType: str, outputType: str) -> None:
#     inputTypes = ["temporal", "utility"]
#     outputTypes = ["sparse", "dense"]

#     inputType = inputType.lower()

#     error = False

#     if inputType not in inputTypes:
#         print("Input type must be one of: " + str(inputTypes))
#         error = True

#     outputType = outputType.lower()
#     if outputType not in outputTypes:
#         print("Output type must be one of: " + str(outputTypes))
#         error = True
    
#     if error:
#         return
    
#     file = csv_file.split(".")
#     parquet_file = file[0] + ".parquet"
#     dict_file = file[0] + ".dict"
    
#     if inputType == "temporal":
#         conversion = {}
#         conNum = 1

#         file = []
#         indexes = []
#         for line in open(csv_file):
#             # first item is the index
#             line = line.strip().split(sep)
#             indexes.append(int(line[0]))
#             # file.append([int(i) for i in line[1:]])
#             temp = []
#             for i in line[1:]:
#                 if i not in conversion:
#                     conversion[i] = conNum
#                     conNum += 1
#                 temp.append(conversion[i])
#             file.append(temp)

#         if outputType == 'dense':
#             sparseList = []
#             indx = []

#             for i in range(len(file)):
#                 for j in range(len(file[i])):
#                     sparseList.append(file[i][j])
#                     indx.append(indexes[i])

#             columns = [str(i) for i in range(1, 2)]
#             df = pd.DataFrame(sparseList, columns=columns, index=indx)
#             df.to_parquet(parquet_file, engine='pyarrow')


#         elif outputType == 'sparse':
#             maxLen = max([len(i) for i in file])

#             for i in range(len(file)):
#                 if len(file[i]) < maxLen:
#                     file[i].extend([-1] * (maxLen - len(file[i])))

#             columns = [str(i) for i in range(1, maxLen+1)]
#             df = pd.DataFrame(file, columns=columns, index=indexes)
#             df.to_parquet(parquet_file, engine='pyarrow')

        

#     elif inputType == "utility":
#         conversion = {}
#         conNum = 1

#         file = []

#         indexes = 1
#         maxLen = 0

#         for line in open(csv_file):
#             line = line.strip().split(":")
#             items = line[0].split(sep)
#             values = line[2].split(sep)
#             values = [float(i) for i in values]
#             temp = []
#             for i in items:
#                 if i not in conversion:
#                     conversion[i] = conNum
#                     conNum += 1
#                 temp.append(conversion[i])
#             temp = [int(i) for i in temp]
#             file.append([indexes, temp, values])
#             maxLen = max(maxLen, len(values))
#             indexes += 1


#         if outputType == 'dense':
#             newFile = []
#             indexes = []

#             for i in range(len(file)):
#                 for j in range(len(file[i][1])):
#                     newFile.append([file[i][1][j], file[i][2][j]])
#                     indexes.append(file[i][0])
            
#             columns = [str(i) for i in range(1, 3)]

#             df = pd.DataFrame(newFile, columns=columns, index=indexes)
#             df.to_parquet(parquet_file, engine='pyarrow')

#         elif outputType == 'sparse':
#             newFile = []
#             indexes = []
#             for i in range(len(file)):
#                 newFile.append([])
#                 for j in range(len(file[i][1])):
#                     newFile[-1].append(file[i][1][j])
#                     newFile[-1].append(file[i][2][j])
#                 indexes.append(file[i][0])

#             columns = [str(i) for i in range(1, (maxLen*2) + 1)]
#             df = pd.DataFrame(newFile, index=indexes, columns=columns)
#             df.to_parquet(parquet_file, engine='pyarrow')

            
#     with open(dict_file, 'w') as f:
#             for key in conversion.keys():
#                 f.write("%s->%s\n"%(key,conversion[key]))


# def parquetFormat(file: str, sep: str, inputType: str, outputType: str) -> None:
#     inputTypes = ["temporal", "utility"]
#     outputTypes = ["sparse", "dense"]

#     inputType = inputType.lower()

#     error = False

#     if inputType not in inputTypes:
#         print("Input type must be one of: " + str(inputTypes))
#         error = True

#     outputType = outputType.lower()
#     if outputType not in outputTypes:
#         print("Output type must be one of: " + str(outputTypes))
#         error = True

#     if error:
#         return
    
#     df = pd.read_parquet(file, engine='pyarrow')

#     if inputType == "temporal":

#         if outputType == "dense":
#             ndf = {}
#             indexes = df.index.tolist()
#             # # ndf = df.values.tolist()
#             for i in range(len(df)):
#                 # ndf[i] = [j for j in ndf[i] if j != -1]
#                 if indexes[i] not in ndf:
#                     ndf[indexes[i]] = []
#                 ndf[indexes[i]].append([j for j in df.iloc[i] if j != -1])

#             sparseList = []
#             indx = []

#             indexes = list(ndf.keys())
#             filed = [x[0] for x in list(ndf.values())]

#             for i in range(len(filed)):
#                 for j in range(len(filed[i])):
#                     sparseList.append(filed[i][j])
#                     indx.append(indexes[i])

#             columns = [str(i) for i in range(1, 2)]
#             df = pd.DataFrame(sparseList, columns=columns, index=indx)
#             df.to_parquet(file, engine='pyarrow')
        
#         if outputType == "sparse":
#             ndf = {}
#             indexes = df.index.tolist()


#             for i in range(len(df)):
#                 if indexes[i] not in ndf:
#                     ndf[indexes[i]] = []
#                 for j in range(len(df.iloc[i])):
#                     # ndf[indexes[i]].append(int(df.iloc[i,j]))
#                     if df.iloc[i,j] != -1:
#                         ndf[indexes[i]].append(int(df.iloc[i,j]))
#                     else:
#                         break

            
#             indexes = list(ndf.keys())
#             nfile = list(ndf.values())

#             maxLen = max([len(i) for i in nfile])

#             for i in range(len(nfile)):
#                 if len(nfile[i]) < maxLen:
#                     nfile[i].extend([-1] * (maxLen - len(nfile[i])))

#             columns = [str(i) for i in range(1, maxLen+1)]
#             df = pd.DataFrame(nfile, columns=columns, index=indexes)
#             df.to_parquet(file, engine='pyarrow')

    
#     if inputType == "utility":
#         ndf = []

#         if outputType == "sparse":
#             ndf = {}

#             indexes = df.index.tolist()

#             for i in range(len(df)):
#                 index = indexes[i]
#                 if index not in ndf:
#                     ndf[index] = []
#                 for j in range(len(df.iloc[i])):
#                     if j % 2 == 0:
#                         ndf[index].append(int(df.iloc[i,j]))
#                     else:
#                         ndf[index].append(float(df.iloc[i,j]))

#             indexes = list(ndf.keys())
#             nfile = list(ndf.values())

#             maxLen = max([len(i) for i in nfile])

            
#             columns = [str(i) for i in range(1, (maxLen+1))]
#             df = pd.DataFrame(nfile, columns=columns, index=indexes)
#             df.to_parquet(file, engine='pyarrow')
            

#         if outputType == "dense":
#             ndf = {}

#             indexes = df.index.tolist()

#             for i in range(len(df)):
#                 index = indexes[i]
#                 if index not in ndf:
#                     ndf[index] = [[],[]]

#                 numbers = df.iloc[i,0::2].dropna().tolist()
#                 values = df.iloc[i,1::2].dropna().tolist()

#                 zipped = list(zip(numbers, values))
#                 for i in range(len(zipped)):
#                     ndf[index][0].append(int(zipped[i][0]))
#                     ndf[index][1].append(float(zipped[i][1]))

#             indexes = list(ndf.keys())
#             nfile = list(ndf.values())

#             newFile = []
#             iindexes = []

#             for i in range(len(nfile)):
#                 for j in range(len(nfile[i][1])):
#                     newFile.append([nfile[i][0][j], nfile[i][1][j]])
#                     iindexes.append(indexes[i])

#             lens = [len(i) for i in newFile]
            
#             columns = [str(i) for i in range(1, 3)]

#             # nums = [i[0] for i in newFile]
#             # vals = [i[1] for i in newFile]

#             df = pd.DataFrame(newFile, columns=columns, index=iindexes)
#             # df = pd.DataFrame([nums, vals], index=iindexes)
#             # print(df)
#             df.to_parquet(file, engine='pyarrow')


import pandas as pd
import os 
import psutil
import time


class CSV2Parquet:
    """
        **About this algorithm**

        :**Description**:  This class is to convert CSV files into Parquet format.

        :**Reference**:

        :**Parameters**:    - **inputFile** (*str*) -- *Path to the input CSV file.*
                            - **outputFile** (*str*) -- *Path to the output Parquet file.*
                            - **sep** (*str*) -- *This variable is used to distinguish items from one another. The default seperator is tab space. However, the users can override their default separator.*

        :**Attributes**:    - **getMemoryUSS** (*float*) -- *Returns the memory used by the process in USS.*
                            - **getMemoryRSS** (*float*) -- *Returns the memory used by the process in RSS.*
                            - **getRuntime()** (*float*) -- *Returns the time taken to execute the conversion.*
                            - **printStats()** -- * Prints statistics about memory usage and runtime.*

        :**Methods**:       - **convert()** -- *Reads the input file, converts it to a Parquet file, and tracks memory usage and runtime.*


        **Execution methods**

        **Terminal command**

        .. code-block:: console

          Format:

          (.venv) $ python3 _CSV2Parquet.py <inputFile> <outputFile> <sep>

          Example Usage:

          (.venv) $ python3 _CSV2Parquet.py sampleDB.csv output.parquet \t


        **Calling from a python program**

        .. code-block:: python

                import PAMI.extras.convert.CSV2Parquet as cp

                inputFile = 'sampleDB.csv'

                sep = "\t"

                outputFile = 'output.parquet'

                obj = cp.CSV2Parquet(inputFile, outputFile, sep)

                obj.convert()

                obj.printStats()


        **Credits**

        The complete program was written by P. Likhitha  and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """
    def __init__(self, inputFile, outputFile, sep):
        self.inputFile = inputFile
        self.outputFile = outputFile
        self.sep = sep

    def convert(self):
        """
        This function converts the input CSV file to a data frame, which is then transformed into a Parquet file.
        """
        self.start = time.time()
        file = []
        maxLen = 0
        with open(self.inputFile, "r") as f:
            for line in f:
                file.append(line.strip().split(self.sep))
                maxLen = max(maxLen, len(file[-1]))

        for i in range(len(file)):
            file[i] += [""] * (maxLen - len(file[i]))

        df = pd.DataFrame(file)

        df.to_parquet(self.outputFile)

        self.end = time.time()

        self.pid = os.getpid()
        process = psutil.Process(self.pid)
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Returns the memory used by the process in USS (Unique Set Size).

        :return: The amount of memory (in bytes) used exclusively by the process
        :rtype: int
        """
        return self.memoryUSS
    
    def getMemoryRSS(self):
        """
        Returns the memory used by the process in RSS (Resident Set Size).

        :return: The total memory (in bytes) used by the process in RAM.
        :rtype: int
        """
        return self.memoryRSS
    
    def getRuntime(self):
        """
        Returns the time taken to complete the CSV to Parquet conversion.

        :return: The runtime of the conversion process in seconds.
        :rtype: float
        """
        return self.end - self.start
    
    def printStats(self):
        """
        Prints the resource usage statistics including memory consumption (USS and RSS) and the runtime.

        :return: Prints memory usage and runtime to the console.
        """
        print("Memory usage (USS):", self.memoryUSS)
        print("Memory usage (RSS):", self.memoryRSS)
        print("Runtime:", self.end - self.start)
    

# file = "Transactional_T10I4D100K.csv"
# sep = "\t"
# outputFile = "output.parquet"
# obj = CSV2Parquet(file, outputFile, sep)
# obj.convert()
# obj.printStats()


class Parquet2CSV:
    """
        **About this algorithm**

        :**Description**:  This class is to convert Parquet format into CSV file.

        :**Reference**:

        :**Parameters**:    - **inputFile** (*str*) -- *Path to the input Parquet file.*
                            - **outputFile** (*str*) -- *Path to the output CSV file.*
                            - **sep** (*str*) -- *This variable is used to distinguish items from one another. The default seperator is tab space. However, the users can override their default separator.*

        :**Attributes**:    - **getMemoryUSS** (*int*) -- *Returns the memory used by the process in USS.*
                            - **getMemoryRSS** (*int*) -- *Returns the memory used by the process in RSS.*
                            - **getRuntime()** (*float*) -- *Returns the time taken to execute the conversion.*
                            - **printStats()** -- * Prints statistics about memory usage and runtime.*

        :**Methods**:       - **convert()** -- *Reads the Parquet file, converts it to a CSV file, and tracks memory usage and runtime.*

        **Execution methods**

        **Terminal command**

        .. code-block:: console

          Format:

          (.venv) $ python3 _CSV2Parquet.py <inputFile> <outputFile> <sep>

          Example Usage:

          (.venv) $ python3 _CSV2Parquet.py output.parquet sampleDB.csv \t


        **Calling from a python program**

        .. code-block:: python

                import PAMI.extras.convert.Parquet2CSV as pc

                inputFile = 'output.parquet'

                sep = "\t"

                outputFile = 'sampleDB.csv'

                obj = pc.Parquet2CSV(inputFile, outputFile, sep)

                obj.convert()

                obj.printStats()


        **Credits**

        The complete program was written by P. Likhitha  and revised by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """
    def __init__(self, inputFile, outputFile, sep):
        self.inputFile = inputFile
        self.outputFile = outputFile
        self.sep = sep

    def convert(self):
        """
        This function converts the input Parquet file into a CSV file where each row is joined by the specified separator and written to the output file.
        """
        self.start = time.time()
        df = pd.read_parquet(self.inputFile)

        with open(self.outputFile, "w") as f:
            for i in range(len(df)):
                f.write(self.sep.join(df.iloc[i]) + "\n")

        self.end = time.time()

        self.pid = os.getpid()
        process = psutil.Process(self.pid)
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss

    def getMemoryUSS(self):
        """
        Returns the memory used by the process in USS (Unique Set Size).

        :return: The amount of memory (in bytes) used exclusively by the process
        :rtype: int
        """
        return self.memoryUSS
    
    def getMemoryRSS(self):
        """
        Returns the memory used by the process in RSS (Resident Set Size).

        :return: The total memory (in bytes) used by the process in RAM.
        :rtype: int
        """
        return self.memoryRSS
    
    def getRuntime(self):
        """
        Returns the time taken to complete the Parquet to CSV conversion.

        :return: The runtime of the conversion process in seconds.
        :rtype: float
        """
        return self.end - self.start
    
    def printStats(self):
        """
        Prints the resource usage statistics including memory consumption (USS and RSS) and the runtime.

        :return: Prints memory usage and runtime to the console.
        """
        print("Memory usage (USS):", self.memoryUSS)
        print("Memory usage (RSS):", self.memoryRSS)
        print("Runtime:", self.end - self.start)

# outputFile = "output.csv"
# obj = Parquet2CSV("output.parquet", outputFile, sep)
# obj.convert()
# obj.printStats()






# imagery2Databases is a code used to creating transactional database by applying threshold
#
#  **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.extras.imageProcessing import imagery2Databases as db
#
#     obj = db.imagery2Databases(detected_objects, 16 )
#
#     obj.save()
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd
import sys


# creating transactional database by applying threshold
class createDatabase:
    """

    :Description:   imagery2Databases is a code used to creating transactional database by applying threshold

    :param  detected_objects: list :
            List data to be processed
    :param  threshold: int :
            It is threshold value of all item


    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.imageProcessing import imagery2Databases as db

            obj = db.imagery2Databases(detected_obj, 16 )

            obj.save(oFile)

    """

    # pass the list of detected objects and specify the minimum probability score an object must maintain.
    def __init__(self, detected_objects: list, threshold: float):
        # initialize data frame to store objects
        self.dataframe = pd.DataFrame(columns=['objects'])
        self.threshold = threshold
        self.itemList = []
        self.probabilityValuesList = []
        self.detected_objects = detected_objects
        self.itemSupport = []
        self.itemSupportSum = []  # pruning the objects having scores less than threshold value
        for objectList in self.detected_objects:
            supportSum = 0
            dataDic = {}
            self.items = []
            self.values = []
            self.supports = []
            for item in objectList:
                supportSum = supportSum + 1
                if item[1] >= self.threshold:
                    if item[0] not in dataDic.keys():
                        dataDic[item[0]] = [item[1]]
                    else:
                        dataDic[item[0]].append(item[1])
            # storing objects,their probabilities and count
            self.items = [item for item in dataDic.keys()]
            self.values = [max(value) for value in dataDic.values()]
            self.supports = [len(value) for value in dataDic.values()]

            self.itemSupportSum.append(supportSum)
            self.itemList.append(self.items)
            self.probabilityValuesList.append(self.values)
            self.itemSupport.append(self.supports)
            self.dataframe.loc[self.dataframe.shape[0], 'objects'] = dataDic.keys()

    def getDataFrame(self) -> pd.DataFrame:
        return self.dataframe

    # This function will save the list of objects found in each image as a transactional database.

    # creating transactional database
    def saveAsTransactionalDB(self, outputFile: str, sep: str) -> None:
        writeFile = open(outputFile, 'w')
        for i in range(len(self.itemList)):
            if self.itemList[i]:
                writeLine = sep.join(map(str, self.itemList[i]))
                writeFile.write(writeLine + '\n')
        writeFile.close()

    # creating temporal database
    def saveAsTemporalDB(self, outputFile: str, sep: str):
        writeFile = open(outputFile, 'w')

        for i in range(len(self.itemList)):
            if self.itemList[i]:
                writeLine = sep.join(map(str, self.itemList[i]))
                writeFile.write(str(i) + sep + writeLine + '\n')

        writeFile.close()

    # creating utility transactional database

    def saveAsUtilityTransactionalDB(self, outputFile: str, sep: str) -> None:
        writeFile = open(outputFile, 'w')
        for i in range(len(self.itemList)):
            if self.itemList[i]:
                writeLine = sep.join(map(str, self.itemList[i]))
                writeLine2 = sep.join(map(str, self.itemSupport[i]))
                writeFile.write(writeLine + ':' + str(self.itemSupportSum[i]) + ':' + writeLine2 + '\n')
        writeFile.close()

    # creating utility temporal database

    def saveAsUtilityTemporalDB(self, outputFile: str, sep: str) -> None:
        writeFile = open(outputFile, 'w')
        for i in range(len(self.itemList)):
            if self.itemList[i]:
                writeLine = sep.join(map(str, self.itemList[i]))
                writeLine2 = sep.join(map(str, self.itemSupport[i]))
                writeFile.write(
                    str(i) + str(sep) + writeLine + ':' + str(self.itemSupportSum[i]) + ':' + writeLine2 + '\n')
        writeFile.close()

    # creating uncertain transactional database

    def saveAsUncertainTransactionalDB(self, outputFile: str, sep: str) -> None:
        writeFile = open(outputFile, 'w')
        for i in range(len(self.itemList)):
            if self.itemList[i]:
                writeLine = sep.join(map(str, self.itemList[i]))
                writeLine2 = sep.join(map(str, self.probabilityValuesList[i]))
                writeFile.write(writeLine + ":1:" + writeLine2 + '\n')
        writeFile.close()

    # creating uncertain Temporal database

    def saveAsUncertainTemporalDB(self, outputFile: str, sep: str) -> None:
        writeFile = open(outputFile, 'w')
        for i in range(len(self.itemList)):
            if self.itemList[i]:
                writeLine = sep.join(map(str, self.itemList[i]))
                writeLine2 = sep.join(map(str, self.probabilityValuesList[i]))
                writeFile.write(str(i) + str(sep) + writeLine + ":1:" + writeLine2 + '\n')
        writeFile.close()

if __name__ == '__main__':
    obj = createDatabase(sys.argv[1], sys.argv[2])
    obj.saveAsUncertainTemporalDB(sys.argv[3])


# CPPG algorithm discovers coverage patterns in a transactional database.
#
# **Importing this algorithm into a python program**
# -------------------------------------------------------
#
#
#             from PAMI.coveragePattern.basic import CPPG as alg
#
#             obj = alg.CPPG(iFile, minRF, minCS, maxOR)
#
#             obj.mine()
#
#             coveragePattern = obj.getPatterns()
#
#             print("Total number of coverage Patterns:", len(coveragePattern))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran
     
"""

from PAMI.coveragePattern.basic import abstract as _ab
import pandas as pd
from typing import List, Dict, Tuple, Set, Union, Any, Generator
from deprecated import deprecated


_maxPer = float()
_minSup = float()
_lno = int()


class CPPG(_ab._coveragePatterns):
    """

    :Description:  CPPG  algorithm discovers coverage patterns in a transactional database.

    :Reference:     Gowtham Srinivas, P.; Krishna Reddy, P.; Trinath, A. V.; Bhargav, S.; Uday Kiran, R. (2015).
                    Mining coverage patterns from transactional databases. Journal of Intelligent Information Systems, 45(3), 423–439.
                    https://link.springer.com/article/10.1007/s10844-014-0318-3

    :param  iFile: str :
                   Name of the Input file to mine complete set of coverage patterns
    :param  oFile: str :
                   Name of the output file to store complete set of coverage patterns
    :param  minRF: str:
                   Controls the minimum number of transactions in which every item must appear in a database.
    :param  minCS: str:
                   Controls the minimum number of transactions in which at least one time within a pattern must appear in a database.
    :param  maxOR: str:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

        startTime : float
          To record the start time of the mining process

        endTime : float
          To record the completion time of the mining process

        finalPatterns : dict
          Storing the complete set of patterns in a dictionary variable

        memoryUSS : float
          To store the total amount of USS memory consumed by the program

        memoryRSS : float
          To store the total amount of RSS memory consumed by the program

        Database : list
          To store the transactions of a database in list


    **Methods to execute code on terminal**
    -------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 CPPG.py <inputFile> <outputFile> <minRF> <minCS> <maxOR> <'\t'>

      Example Usage:

      (.venv) $ python3 CPPG.py sampleTDB.txt patterns.txt 0.4 0.7 0.5 ','

    .. note:: minSup will be considered in percentage of database transactions


    **Importing this algorithm into a python program**
    --------------------------------------------------

    .. code-block:: python

            from PAMI.coveragePattern.basic import CPPG as alg

            obj = alg.CPPG(iFile, minRF, minCS, maxOR)

            obj.mine()

            coveragePattern = obj.getPatterns()

            print("Total number of coverage Patterns:", len(coveragePattern))

            obj.save(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits:**
    -------------------------
             The complete program was written by P.Likhitha  under the supervision of Professor Rage Uday Kiran.

    """
    _startTime = float()
    _endTime = float()
    _minRF = str()
    _maxOR = str()
    _minCS = str()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _rank = {}
    _rankedUp = {}
    _lno = 0

    def _creatingItemSets(self) -> None:
        """
            Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            data, ts = [], []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'TS' in i:
                ts = self._iFile['TS'].tolist()
            if 'Transactions' in i:
                data = self._iFile['Transactions'].tolist()
            for i in range(len(data)):
                tr = [ts[i][0]]
                tr = tr + data[i]
                self._Database.append(tr)

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self._Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()


    def _coverageOneItem(self) -> Tuple[Dict[str, List[int]], List[str]]:
        """ Calculates the support of each item in the database and assign ranks to the items
            by decreasing support and returns the frequent items list

            :returns: return the one-length periodic frequent patterns
            :rtype: tuple
        """
        data = {}
        count = 0
        for tr in self._Database:
            count += 1
            for i in range(len(tr)):
                if tr[i] not in data:
                    data[tr[i]] = [count]
                else:
                    data[tr[i]].append(count)
        data = {k: v for k, v in data.items() if len(v)/len(self._Database) >= self._minRF}
        pfList = [i for i in sorted(data, key=lambda k: len(data[k]), reverse=True)]
        return data, pfList

    def _updateDatabases(self, dict1: Dict[str, List[str]]) -> List[List[str]]:
        """ Remove the items which are not frequent from database and updates the database with rank of items

            :param dict1: frequent items with support
            :type dict1: dict
            :return: Sorted and updated transactions
            :rtype: list
            """
        list2 = []
        for tr in self._Database:
            list1 = []
            for i in range(len(tr)):
                if tr[i] in dict1:
                    list1.append(tr[i])
            list2.append([i for i in dict1 if i in list1])
        return list2

    def _buildProjectedDatabase(self, data: List[List[str]], info: List[str]) -> Dict[str, List[List[str]]]:
        """ To construct the projected database for each prefix
        :param data: list of transactions with support per prefix
        :type data: list
        :param info: informatoin on list of transactions with support per prefix
        :type info: str
        :return: projected data
        :rtype: dict
        """
        proData = {}
        for i in range(len(info)):
            prefix = info[i+1:]
            proData[info[i]] = []
            for j in data:
                te = []
                if info[i] not in j:
                    for k in j:
                        if k in prefix:
                            te.append(k)
                if len(te) > 0:
                    proData[info[i]].append(te)
        for x, y in proData.items():
            print(x, y)
        return proData

    def _generateFrequentPatterns(self,  uniqueItems: List[str]) -> None:
        """It will generate the combinations of frequent items

        :param uniqueItems :it represents the items with their respective transaction identifiers

        :type uniqueItems: list

        :return: returning transaction dictionary

        :rtype: dict
        """
        new_freqList = []
        for i in range(0, len(uniqueItems)):
            item1 = uniqueItems[i]
            i1_list = item1.split()
            for j in range(i + 1, len(uniqueItems)):
                item2 = uniqueItems[j]
                i2_list = item2.split()
                if i1_list[:-1] == i2_list[:-1]:
                    interSet = set(self._finalPatterns[item1]).intersection(set(self._finalPatterns[item2]))
                    union = set(self._finalPatterns[item1]).union(set(self._finalPatterns[item2]))
                    if len(union)/len(self._Database) >= self._minCS and len(interSet)/len(self._finalPatterns[item1]) <= self._maxOR:
                        newKey = item1 + " " + i2_list[-1]
                        self._finalPatterns[newKey] = interSet
                        new_freqList.append(newKey)
                else:
                    break

        if len(new_freqList) > 0:
            self._generateFrequentPatterns(new_freqList)

    def _savePeriodic(self, itemSet: List[str]) -> str:
        """ To convert the ranks of items in to their original item names

            :param itemSet: frequent patterns

            :type itemSet: list

            :return: frequent pattern with original item names

            :rtype: string
        """
        t1 = str()
        for i in itemSet:
            t1 = t1 + self._rankedUp[i] + "\t"
        return t1

    def _convert(self, value: Union[int, float, str]) -> Union[int, float]:
        """
        To convert the given user specified value

        :param value: user specified value

        :type value: Union[int, float, str]

        :return: converted value

        :rtype: Union[int, float]
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = value
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = value
            else:
                value = int(value)
        return value

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self) -> None:
        """ Mining process will start from this function
        """
        self.mine()

    def mine(self) -> None:
        """ Mining process will start from this function
        """

        #global _minSup, _maxPer, _lno
        self._startTime = _ab._time.time()
        if self._iFile is None:
            raise Exception("Please enter the file path or file name:")
        if self._minRF is None:
            raise Exception("Please enter the Relative Frequency")
        if self._maxOR is None:
            raise Exception("Please enter the Overlap Ratio")
        if self._minCS is None:
            raise Exception("Please enter the Coverage Ratio")
        self._creatingItemSets()
        self._minRF = self._convert(self._minRF)
        self._maxOR = self._convert(self._maxOR)
        self._minCS = self._convert(self._minCS)
        if self._minRF > len(self._Database) or self._minCS > len(self._Database) or self._maxOR > len(self._Database):
            raise Exception("Please enter the constraints in range between 0 to 1")
        generatedItems, pfList = self._coverageOneItem()
        self._finalPatterns = {k: v for k, v in generatedItems.items()}
        updatedDatabases = self._updateDatabases(pfList)
        proData = self._buildProjectedDatabase(updatedDatabases, pfList)
        for x, y in proData.items():
            uniqueItems = [x]
            for i in y:
                for j in i:
                    if j not in uniqueItems:
                        uniqueItems.append(j)
            self._generateFrequentPatterns(uniqueItems)
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Coverage patterns were generated successfully using CPPG algorithm ")

    def getMemoryUSS(self) -> float:
        """Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self) -> float:
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self) -> float:
        """Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self) -> pd.DataFrame:
        """Storing final periodic-frequent patterns in a dataframe

        :return: returning periodic-frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b[0], b[1]])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support', 'Periodicity'])
        return dataFrame

    def save(self, outFile: str) -> None:
        """Complete set of periodic-frequent patterns will be loaded in to an output file

        :param outFile: name of the outputfile
        :type outFile: file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x.strip() + ":" + str(len(y))
            writer.write("%s \n" % s1)

    def getPatterns(self) -> Dict[str, List[int]]:
        """ Function to send the set of periodic-frequent patterns after completion of the mining process

        :return: returning periodic-frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self) -> None:
        """
           Function used to print the result
        """
        print("Total number of Coverage Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:",  self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 6 or len(_ab._sys.argv) == 7:
        if len(_ab._sys.argv) == 7:
            _ap = CPPG(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5], _ab._sys.argv[6])
        if len(_ab._sys.argv) == 6:
            _ap = CPPG(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4], _ab._sys.argv[5])
        _ap.mine()
        _ap.mine()
        print("Total number of Coverage Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:",  _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



# generateTransactionalDatabase is a code used to convert the database into Temporal database.
#
#  **Importing this algorithm into a python program**
#  --------------------------------------------------------
#     from PAMI.extras.generateDatabase import generateTransactionalDatabase as db
#     obj = db(10, 5, 10)
#     obj.create()
#     obj.save('db.txt')
#     print(obj.getTransactions()) to get the transactional database as a pandas dataframe

# **Running the code from the command line**
# --------------------------------------------------------
#     python generateDatabase.py 10 5 10 db.txt
#     cat db.txt
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import numpy as np
import pandas as pd
import sys


class generateSpatioTransactionalDatabase:
    """
    :Description Generate a transactional database with the given number of lines, average number of items per line, and total number of items

    :Attributes:
    numLines: int  
        - number of lines
    avgItemsPerLine: int 
        - average number of items per line
    numItems: int 
        - total number of items

    :Methods:
        create: 
            Generate the transactional database
        save: 
            Save the transactional database to a file
        getTransactions: 
            Get the transactional database



    
    """
    def getPoint(self, x1, y1, x2, y2):

        return (np.random.randint(x1, x2), np.random.randint(y1, y2))

    def __init__(self, numLines, avgItemsPerLine, numItems, x1, y1, x2, y2) -> None:
        """
        Initialize the transactional database with the given parameters

        Parameters:
        numLines: int - number of lines
        avgItemsPerLine: int - average number of items per line
        numItems: int - total number of items
        """

        self.numLines = numLines
        self.avgItemsPerLine = avgItemsPerLine
        self.numItems = numItems
        self.db = []
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

        numPoints = (x2 - x1) * (y2 - y1)
        if numItems > numPoints:
            raise ValueError("Number of points is less than the number of lines * average items per line")
        
        self.itemPoint = {}
        usedPoints = set()

        for i in range(1, numItems + 1):
            # self.itemPoint[i] = (np.random.randint(x1, x2), np.random.randint(y1, y2))
            point = self.getPoint(x1, y1, x2, y2)
            while point in usedPoints:
                point = self.getPoint(x1, y1, x2, y2)
            self.itemPoint[i] = point
    
    def tuning(self, array, sumRes) -> list:
        """
        Tune the array so that the sum of the values is equal to sumRes

        :param array: list of values

        :type array: list

        :param sumRes: the sum of the values in the array to be tuned

        :type sumRes: int

        :return: list of values with the tuned values and the sum of the values in the array to be tuned and sumRes is equal to sumRes

        :rtype: list
        """

        while np.sum(array) != sumRes:
            # get index of largest value
            randIndex = np.random.randint(0, len(array))
            # if sum is too large, decrease the largest value
            if np.sum(array) > sumRes:
                array[randIndex] -= 1
            # if sum is too small, increase the smallest value
            else:
                minIndex = np.argmin(array)
                array[randIndex] += 1
        return array
        

    def generateArray(self, nums, avg, maxItems) -> list:
        """
        Generate a random array of length n whose values average to m

        :param nums: number of values

        :type nums: list

        :param avg: average value

        :type avg: float

        :param maxItems: maximum value

        :type maxItems: int

        :return: random array

        :rtype: list
        """

        # generate n random values
        values = np.random.randint(1, maxItems, nums)

        sumRes = nums * avg

        self.tuning(values, sumRes)

        # if any value is less than 1, increase it and tune the array again
        while np.any(values < 1):
            for i in range(nums):
                if values[i] < 1:
                    values[i] += 1
            self.tuning(values, sumRes)

        while np.any(values > maxItems):
            for i in range(nums):
                if values[i] > maxItems:
                    values[i] -= 1
            self.tuning(values, sumRes)


        # if all values are same then randomly increase one value and decrease another
        while np.all(values == values[0]):
            values[np.random.randint(0, nums)] += 1
            self.tuning(values, sumRes)

        return values

    def create(self) -> None:
        """
        Generate the transactional database
        :return: None
        """
        db = set()

        values = self.generateArray(self.numLines, self.avgItemsPerLine, self.numItems)

        for value in values:
            line = np.random.choice(range(1, self.numItems + 1), value, replace=False)
            nline = [self.itemPoint[i] for i in line]
            # print(line, nline)
            # for i in range(len(line)):
            #     print(line[i], self.itemPoint[line[i]])
            #     line[i] = self.itemPoint[line[i]]
            self.db.append(nline)
            # self.db.append(line)

    def save(self, filename) -> None:
        """
        Save the transactional database to a file

        :param filename: name of the file

        :type filename: str

        :return: None
        """

        with open(filename, 'w') as f:
            for line in self.db:
                # f.write(','.join(map(str, line)) + '\n')
                line = list(map(str, line))
                f.write(','.join(line) + '\n')

    def getTransactions(self) -> pd.DataFrame:
        """
        Get the transactional database

        :return: the transactional database

        :rtype: pd.DataFrame
        """
        df = pd.DataFrame(['\t'.join(map(str, line)) for line in self.db])
        return df
        

if __name__ == "__main__":
    # test the class
    db = generateSpatioTransactionalDatabase(10, 5, 10, 1,5,5,10)
    db.create()
    db.save('db.txt')
    print(db.getTransactions())

    obj = generateSpatioTransactionalDatabase(sys.argv[1], sys.argv[2], sys.argv[3])
    obj.create()
    obj.save(sys.argv[4])
    # print(obj.getTransactions())
    


import random as _rd
import sys as _sys
import time
import os
import psutil
import numpy as np
import tqdm


class GeoReferentialTemporalDatabase:
    """
    This class create synthetic geo-referential temporal database.

    :Attribute:

        totalTransactions : int
            No of transactions
        noOfItems : int or float
            No of items
        avgTransactionLength : str
            The length of average transaction
        outputFile: str
            Name of the output file.

    :Methods:

        GeoReferentialTemporalDatabase(outputFile)
            Create geo-referential temporal database and store into outputFile

    **Credits:**
    ---------------
             The complete program was written by  P.Likhitha   under the supervision of Professor Rage Uday Kiran.

    """

    def __init__(
            self,
            databaseSize: int,
            avgItemsPerTransaction: int,
            numItems: int,
            x1: int,
            y1: int,
            x2: int,
            y2: int,
            sep: str = '\t',
            occurrenceProbabilityOfSameTimestamp: float = 0,
            occurrenceProbabilityToSkipSubsequentTimestamp: float = 0,
    ) -> None:
        self.databaseSize = databaseSize
        self.avgItemsPerTransaction = avgItemsPerTransaction
        self.numItems = numItems
        self.db = []
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.seperator = sep
        self.occurrenceProbabilityOfSameTimestamp = occurrenceProbabilityOfSameTimestamp
        self.occurrenceProbabilityToSkipSubsequentTimestamp = occurrenceProbabilityToSkipSubsequentTimestamp
        self._startTime = float()
        self._endTime = float()
        self._memoryUSS = float()
        self._memoryRSS = float()
        if numItems > ((x2 - x1) * (y2 - y1)):
            raise ValueError("Number of points is less than the number of lines * average items per line")

        self.itemPoint = {}
        usedPoints = set()

        for i in range(1, numItems + 1):
            # self.itemPoint[i] = (np.random.randint(x1, x2), np.random.randint(y1, y2))
            point = self.getPoint(x1, y1, x2, y2)
            while point in usedPoints:
                point = self.getPoint(x1, y1, x2, y2)
            self.itemPoint[i] = point

    def getPoint(self, x1, y1, x2, y2):

        return (np.random.randint(x1, x2), np.random.randint(y1, y2))

    def performCoinFlip(self, probability: float) -> bool:
        """
        Perform a coin flip with the given probability.

        :param probability: Probability of the coin landing heads (i.e., the event occurring).
        :return: True if the coin lands heads, False otherwise.
        """
        result = np.random.choice([0, 1], p=[1 - probability, probability])
        return result == 1

    def tuning(self, array, sumRes) -> list:
        """
        Tune the array so that the sum of the values is equal to sumRes

        :param array: list of values

        :type array: list

        :param sumRes: the sum of the values in the array to be tuned

        :type sumRes: int

        :return: list of values with the tuned values and the sum of the values in the array to be tuned and sumRes is equal to sumRes

        :rtype: list
        """

        while np.sum(array) != sumRes:
            # get index of largest value
            randIndex = np.random.randint(0, len(array))
            # if sum is too large, decrease the largest value
            if np.sum(array) > sumRes:
                array[randIndex] -= 1
            # if sum is too small, increase the smallest value
            else:
                minIndex = np.argmin(array)
                array[randIndex] += 1
        return array

    def generateArray(self, nums, avg, maxItems) -> list:
        """
        Generate a random array of length n whose values average to m

        :param nums: number of values

        :type nums: list

        :param avg: average value

        :type avg: float

        :param maxItems: maximum value

        :type maxItems: int

        :return: random array

        :rtype: list
        """

        # generate n random values
        values = np.random.randint(1, maxItems, nums)

        sumRes = nums * avg

        self.tuning(values, sumRes)

        # if any value is less than 1, increase it and tune the array again
        while np.any(values < 1):
            for i in range(nums):
                if values[i] < 1:
                    values[i] += 1
            self.tuning(values, sumRes)

        while np.any(values > maxItems):
            for i in range(nums):
                if values[i] > maxItems:
                    values[i] -= 1
            self.tuning(values, sumRes)

        # if all values are same then randomly increase one value and decrease another
        while np.all(values == values[0]):
            values[np.random.randint(0, nums)] += 1
            self.tuning(values, sumRes)

        return values

    def create(self) -> None:
        """
        Generate the Temporal database
        :return: None
        """
        self._startTime = time.time()
        db = set()

        sumRes = self.databaseSize * self.avgItemsPerTransaction  # Total number of items

        for i in range(self.databaseSize):
            # Determine the timestamp
            if self.performCoinFlip(self.occurrenceProbabilityOfSameTimestamp):
                timestamp = self.current_timestamp
            else:
                if self.performCoinFlip(self.occurrenceProbabilityToSkipSubsequentTimestamp):
                    self.current_timestamp += 2
                else:
                    self.current_timestamp += 1
                timestamp = self.current_timestamp

            self.db.append([timestamp])  # Start the transaction with the timestamp

            lineSize.append([i, 0])  # Initialize lineSize with 0 for each transaction

        # Adjust lineSize to ensure sum of sizes equals sumRes
        lineSize = self.tuning(lineSize, sumRes)

        # For each transaction, generate items
        for i in tqdm.tqdm(range(len(lineSize))):
            transaction_index = lineSize[i][0]
            num_items = lineSize[i][1]

            if num_items > self.numItems:
                raise ValueError(
                    "Error: Either increase numItems or decrease avgItemsPerTransaction or modify percentage")
            items = np.random.choice(range(1, self.numItems + 1), num_items, replace=False)
            self.db[transaction_index].extend(items)

        self._runTime = time.time() - start
        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss

    def save(self,filename, sep='\t') -> None:
        """
        Save the Temporal database to a file

        :param filename: name of the file

        :type filename: str

        :param sep: seperator for the items

        :type sep: str

        :return: None
        """

        with open(filename, 'w') as f:
            for line in self.db:
                # f.write(','.join(map(str, line)) + '\n')
                line = list(map(str, line))
                f.write(sep.join(line) + '\n')
    def getTransactions(self) -> pd.DataFrame:
        """
        Get the Temporal database

        :return: the Temporal database

        :rtype: pd.DataFrame
        """
        df = pd.DataFrame(['\t'.join(map(str, line)) for line in self.db], columns=['Transactions'])
        return df


    def getRuntime(self) -> float:
        """
        Get the runtime of the transactional database

        :return: the runtime of the transactional database


        :rtype: float
        """
        return self._endTime - self._startTime

    def getMemoryUSS(self) -> float:

        return self._memoryUSS

    def getMemoryRSS(self) -> float:

        return self._memoryRSS
# if __name__ == "__main__":
#     _ap = str()
#     _ap = createSyntheticGeoreferentialTemporal(100000, 870, 10)
#     _ap.GeoreferentialTemporalDatabase("T10_geo_temp.txt")
# else:
#     print("Error! The number of input parameters do not match the total number of parameters provided")



# findNeighboursUsingGeodesic is a code used to create a neighbourhood file using Geodesic distance.
#
#  **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.extras.neighbours import findNeighboursUsingGeodesic as db
#
#     obj = db.findNeighboursUsingGeodesic(iFile, oFile, 10, "\t")
#
#     obj.save()
#


__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import re
from geopy.distance import geodesic
import time
import sys
import psutil,os,tqdm
import pandas as pd


class FindNeighboursUsingGeodesic:
    """
    This class create a neighbourhood file using Geodesic distance.

    :Attribute:

        :param iFile : file
            Input file name or path of the input file
        :param maxDist : float
            The user can specify maxDist in Km(Kilometers).
            This program find pairs of values whose Geodesic distance is less than or equal to maxDistace
            and store the pairs.
        :param  sep: str :
                    This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Methods:

        mine()
            find and store the pairs of values whose Geodesic distance is less than or equal to maxDistace.
        getFileName()
            This function returns output file name.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.neighbours import findNeighboursUsingGeodesic as db

             obj = db.findNeighboursUsingGeodesic(iFile, oFile, 10, "\t")

            obj.save()
    """

    def __init__(self, iFile: str, maxDist: float, sep='\t',DBtype="temp"):
        self.iFile = iFile
        self.maxGeodesicDistance = maxDist
        self.seperator = sep
        self.result = {}
        self.DBtype = DBtype
        self._startTime = float()
        self._endTime = float()
        self._memoryUSS = float()
        self._memoryRSS = float()



    def create(self) -> None:
        self._startTime = time.time()
        coordinates = []
        with open(self.iFile, "r") as f:
            if self.DBtype == "temp":
                for line in f:
                    l = line.rstrip().split(self.seperator)
                    for i in l[1:]:
                        i = re.sub(r'[^0-9. ]', '', i)
                        if i not in coordinates:
                            coordinates.append(i.rstrip().split(' '))
            else:
                for line in f:
                    l = line.rstrip().split(self.seperator)
                    for i in l:
                        i = re.sub(r'[^0-9. ]', '', i)
                        if i not in coordinates:
                            coordinates.append(i.rstrip().split(' '))
        for i in tqdm.tqdm(range(len(coordinates))):
            for j in range(len(coordinates - i - 1)):
                    j = j + i + 1
                    firstCoordinate = coordinates[i]
                    secondCoordinate = coordinates[j]
                    long1 = float(firstCoordinate[0])
                    lat1 = float(firstCoordinate[1])
                    long2 = float(secondCoordinate[0])
                    lat2 = float(secondCoordinate[1])

                    dist = geodesic((lat1, long1), (lat2, long2)).kilometers

                    if dist <= float(self.maxGeodesicDistance):
                        self.result[tuple(firstCoordinate)] = self.result.get(tuple(firstCoordinate), [])
                        self.result[tuple(firstCoordinate)].append(secondCoordinate)
                        self.result[tuple(secondCoordinate)] = self.result.get(tuple(secondCoordinate), [])
                        self.result[tuple(secondCoordinate)].append(firstCoordinate)
        self._endTime = time.time()

    def save(self, oFile: str) -> None:
        with open(oFile, "w+") as f:
            for i in self.result:
                string = "Point(" + i[0] + " " + i[1] + ")" + self.seperator
                f.write(string)
                for j in self.result[i]:
                    string = "Point(" + j[0] + " " + j[1] + ")" + self.seperator
                    f.write(string)
                f.write("\n")

    def getNeighboringInformation(self):
        df = pd.DataFrame(['\t'.join(map(str, line)) for line in self.result], columns=['Neighbors'])
        return df
    def getRuntime(self) -> float:
        """
        Get the runtime of the transactional database

        :return: the runtime of the transactional database


        :rtype: float
        """
        return self._endTime - self._startTime

    def getMemoryUSS(self) -> float:

        process = psutil.Process(os.getpid())
        self._memoryUSS = process.memory_full_info().uss
        return self._memoryUSS

    def getMemoryRSS(self) -> float:

        process = psutil.Process(os.getpid())
        self._memoryRSS = process.memory_info().rss
        return self._memoryRSS


if __name__ == "__main__":
    obj = FindNeighboursUsingGeodesic(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])



#
#    from PAMI.extras.graph import graphTransactionsExtras as gte
#
#    gte = gte.GraphTransactionExtras("path/to/file")
#
#    graph = gte.getGraphById(graphId)
#
#    gte.plotGraphById(graphId)
#
#    allGraphs = gte.getAllGraphs()


import matplotlib.pyplot as plt
import networkx as nx
from PAMI.subgraphMining.basic import abstract as _ab


class GraphTransactionExtras:
    def __init__(self, filePath):
        """
        Initialize the GraphDatabase with the given file path.
        :param filePath: Path to the file containing graph transactions.
        """
        self.filePath = filePath
        self.graphs = self.readGraphs()

    def readGraphs(self):
        """
        Reads graph data from the specified file and constructs a list of graphs.
        :return: A list of graph objects.
        """
        graphs = []
        with open(self.filePath, 'r') as br:
            vMap = {}
            gId = None

            for line in br:
                line = line.strip()

                if line.startswith("t"):
                    if vMap:  # If vMap is not empty, it means a graph was read
                        graphs.append(_ab.Graph(gId, vMap))
                        vMap = {}  # Reset for the next graph

                    gId = int(line.split(" ")[2])

                elif line.startswith("v"):
                    items = line.split(" ")
                    vId = int(items[1])
                    vLabel = int(items[2])
                    vMap[vId] = _ab.Vertex(vId, vLabel)

                elif line.startswith("e"):
                    items = line.split(" ")
                    v1 = int(items[1])
                    v2 = int(items[2])
                    eLabel = int(items[3])
                    e = _ab.Edge(v1, v2, eLabel)
                    vMap[v1].addEdge(e)
                    vMap[v2].addEdge(e)

            if vMap:
                graphs.append(_ab.Graph(gId, vMap))

        return graphs

    def getGraphById(self, graphId):
        """
        Retrieves a specific graph by its ID.
        :param graph_id: The ID of the graph to retrieve.
        :return: The graph object with the specified ID or None if not found.
        """
        for graph in self.graphs:
            if graph.getId() == graphId:
                return graph
        return None

    def getAllGraphs(self):
        """
        Retrieves all graphs in the database.
        :return: A list of all graph objects.
        """
        return self.graphs

    def plotGraphById(self, graphId):
        """
        Plots a graph by its ID using matplotlib and networkx.
        :param graphId: The ID of the graph to plot.
        """
        graph = self.getGraphById(graphId)
        if not graph:
            print(f"Graph with ID {graphId} not found.")
            return

        G = nx.Graph()

        for vertex in graph.getAllVertices():
            G.add_node(vertex.getId(), label=vertex.getLabel())
            for edge in vertex.getEdgeList():
                G.add_edge(edge.v1, edge.v2, label=edge.getEdgeLabel())

        pos = nx.spring_layout(G)  # You can use other layouts like circular_layout, shell_layout, etc.
        labels = nx.get_node_attributes(G, 'label')
        edge_labels = nx.get_edge_attributes(G, 'label')

        # Draw nodes
        nx.draw_networkx_nodes(G, pos, node_size=700, node_color='lightblue')

        # Draw edges
        nx.draw_networkx_edges(G, pos, width=2)

        # Draw node labels
        nx.draw_networkx_labels(G, pos, labels, font_size=12, font_color='black')

        # Draw edge labels
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_color='red', font_size=10)

        plt.title(f"Graph ID {graphId}")
        plt.axis('off')  # Hide the axes
        plt.show()




# gSpan is a subgraph mining algorithm that uses DFS and DFS codes to mine subgraphs
#
# **Importing this algorithm into a python program**
#
#             from PAMI.subgraphMining.basic import gspan as alg
#
#             obj = alg.GSpan(iFile, minSupport)
#
#             obj.mine()
#
#             obj.run()
#
#             frequentGraphs = obj.getFrequentSubgraphs()
#
#             memUSS = obj.getMemoryUSS()
#
#             obj.save(oFile)
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from .edge import Edge
from .vertex import Vertex

class Graph:
    emptyVertexList = []
    emptyIntegerArray = []

    def __init__(self, id, vMap=None, dfsCode=None):
        """
        The `__init__` function initializes a graph object with optional parameters for vertex mapping and
        DFS code.
        """
        self.vMap = {}
        self.id = id
        if vMap is not None:
            self.vMap = vMap
        elif dfsCode is not None:
            for ee in dfsCode.getEeList():
                v1, v2, v1Label, v2Label, eLabel = ee.v1, ee.v2, ee.vLabel1, ee.vLabel2, ee.edgeLabel
                
                e = Edge(v1, v2, eLabel)
                if v1 not in self.vMap:
                    self.vMap[v1] = Vertex(v1, v1Label)
                if v2 not in self.vMap:
                    self.vMap[v2] = Vertex(v2, v2Label)

                self.vMap[v1].addEdge(e)
                self.vMap[v2].addEdge(e)

            self.id = -1

        self.vertices = []
        self.neighborCache = {}
        self.mapLabelToVertexIds = {}
        self.edgeCount = 0

        self.precalculateVertexList()
        self.precalculateVertexNeighbors()
        self.precalculateLabelsToVertices()

    def getId(self):
        return self.id

    def removeInfrequentLabel(self, label):
        """
        The function removes vertices with a specific label from the graph and updates the edges accordingly.
        """
        toRemove = [key for key, vertex in self.vMap.items() if vertex.getLabel() == label]
        for key in toRemove:
            del self.vMap[key]

        for vertex in self.vMap.values():
            edgesToRemove = [edge for edge in vertex.getEdgeList() 
                               if edge.v1 not in self.vMap or edge.v2 not in self.vMap]

            for edge in edgesToRemove:
                vertex.getEdgeList().remove(edge)

    def precalculateVertexNeighbors(self):
        """
        The function precalculates the neighbors of each vertex in a graph and stores them in a cache.
        """
        self.neighborCache = {}
        self.edgeCount = 0

        for vertexId, vertex in self.vMap.items():
            neighbors = []

            for edge in vertex.getEdgeList():
                neighborVertex = self.vMap[edge.another(vertexId)]
                neighbors.append(neighborVertex)

            neighbors.sort(key=lambda x: x.id)

            self.neighborCache[vertexId] = neighbors
            self.edgeCount += len(neighbors)

        self.edgeCount //= 2    
    
    def precalculateVertexList(self):
        """
        The function precalculateVertexList creates a list of vertices by iterating through a dictionary of
        vertices.
        """
        self.vertices = []
        for _, vertex in self.vMap.items():
            self.vertices.append(vertex)

    def precalculateLabelsToVertices(self):
        """
        This function precalculates and stores mappings of vertex labels to their corresponding vertex IDs.
        """
        self.mapLabelToVertexIds = {}
        for vertex in self.vertices:
            label = vertex.getLabel()
            if label not in self.mapLabelToVertexIds:
                sameIds = [v.getId() for v in self.vertices if v.getLabel() == label]
                self.mapLabelToVertexIds[label] = sameIds

    def findAllWithLabel(self, targetLabel):
        if targetLabel in self.mapLabelToVertexIds:
            return self.mapLabelToVertexIds[targetLabel]
        else:
            return []
        
    def getAllNeighbors(self, v):
        try:
            neighbors = self.neighborCache[v]
        except KeyError:
            neighbors = []
        return neighbors
    
    def getVLabel(self, v):
        return self.vMap[v].getLabel()
    
    def getEdgeLabel(self, v1, v2):
        for e in self.vMap.get(v1).getEdgeList():
            if e.v1 == v1 and e.v2 == v2:
                return e.getEdgeLabel()
            elif e.v1 == v2 and e.v2 == v1:
                return e.getEdgeLabel()
        return -1

    def getEdge(self, v1, v2):
        for e in self.vMap.get(v1).getEdgeList():
            if e.v1 == v1 and e.v2 == v2:
                return e
            elif e.v1 == v2 and e.v2 == v1:
                return e
        return None
    
    def getNonPrecalculatedAllVertices(self):
        return list(self.vMap.values())
    
    def isNeighboring(self, v1, v2):
        neighborsOfV1 = self.neighborCache.get(v1, [])
        low = 0
        high = len(neighborsOfV1) - 1

        while high >= low:
            middle = (low + high) // 2
            val = neighborsOfV1[middle].id
            if val == v2:
                return True
            if val < v2:
                low = middle + 1
            if val > v2:
                high = middle - 1
        return False

    def getAllVertices(self):
        return self.vertices
    
    def getEdgeCount(self):
        return self.edgeCount



# Top K Spatial High Utility ItemSet Mining (TKSHUIM) aims to discover Top-K Spatial High Utility Itemsets
# (TKSHUIs) in a spatioTemporal database
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.highUtilitySpatialPattern.topk import TKSHUIM as alg
#
#             obj=alg.TKSHUIM("input.txt","Neighbours.txt",35)
#
#             obj.mine()
#
#             Patterns = obj.getPatterns()
#
#             print("Total number of  Patterns:", len(Patterns))
#
#             obj.save("output")
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.highUtilitySpatialPattern.topk.abstract import *
from functools import cmp_to_key
import heapq
from deprecated import deprecated

class Transaction:
    """
    A class to store Transaction of a database

    :Attributes:

        items: list
            A list of items in transaction 
        utilities: list
            A list of utilites of items in transaction
        transactionUtility: int
            represent total sum of all utilities in the database
        pmus: list
            represent the pmu (probable maximum utility) of each element in the transaction
        prefixutility:
            prefix Utility values of item
        offset:
            an offset pointer, used by projected transactions

    :Methods:

        projectedTransaction(offsetE):
            A method to create new Transaction from existing till offsetE
        getItems():
            return items in transaction
        getUtilities():
            return utilities in transaction
        getPmus():
            return pmus in transaction
        getLastPosition():
            return last position in a transaction
        removeUnpromisingItems():
            A method to remove items with low Utility than minUtil
        insertionSort():
            A method to sort all items in the transaction
    """
    offset = 0
    prefixUtility = 0
    
    def __init__(self, items, utilities, transactionUtility, pmus=None):
        self.items = items
        self.utilities = utilities
        self.transactionUtility = transactionUtility
        if pmus is not None:
            self.pmus = pmus

    def projectTransaction(self, offsetE):
        """
        A method to create new Transaction from existing till offsetE

        :param offsetE: an offset over the original transaction for projecting the transaction
        :type offsetE: int
        """
        new_transaction = Transaction(self.items, self.utilities, self.transactionUtility)
        utilityE = self.utilities[offsetE]
        new_transaction.prefixUtility = self.prefixUtility + utilityE
        new_transaction.transactionUtility = self.transactionUtility - utilityE
        for i in range(self.offset, offsetE):
            new_transaction.transactionUtility -= self.utilities[i]
        new_transaction.offset = offsetE + 1
        return new_transaction

    def getItems(self):
        """
        A method to return items in transaction
        """
        return self.items

    def getPmus(self):
        """
        A method to return pmus in transaction
        """
        return self.pmus

    def getUtilities(self):
        """
        A method to return utilities in transaction
        """
        return self.utilities

    # get the last position in this transaction
    def getLastPosition(self):
        """
        A method to return last position in a transaction
        """
        return len(self.items) - 1

    def removeUnpromisingItems(self, oldNamesToNewNames):
        """
        A method to remove items with low Utility than minUtil

        :param oldNamesToNewNames: A map represent old namses to new names
        :type oldNamesToNewNames: map
        """
        tempItems = []
        tempUtilities = []
        for idx, item in enumerate(self.items):
            if item in oldNamesToNewNames:
                tempItems.append(oldNamesToNewNames[item])
                tempUtilities.append(self.utilities[idx])
            else:
                self.transactionUtility -= self.utilities[idx]
        self.items = tempItems
        self.utilities = tempUtilities
        self.insertionSort()

    def insertionSort(self):
        """
        A method to sort items in order
        """
        for i in range(1, len(self.items)):
            key = self.items[i]
            utilityJ = self.utilities[i]
            j = i - 1
            while j >= 0 and key < self.items[j]:
                self.items[j + 1] = self.items[j]
                self.utilities[j + 1] = self.utilities[j]
                j -= 1
            self.items[j + 1] = key
            self.utilities[j + 1] = utilityJ


class Dataset:
    """
    A class represent the list of transactions in this dataset

    :Attributes:

    transactions:
        the list of transactions in this dataset
    maxItem:
        the largest item name
        
    :methods:

        createTransaction(line):
            Create a transaction object from a line from the input file
        getMaxItem():
            return Maximum Item
        getTransactions():
            return transactions in database

    """
    transactions = []
    maxItem = 0
    
    def __init__(self, datasetpath, sep):
        self.strToint = {}
        self.intTostr = {}
        self.cnt = 1
        self.sep = sep
        with open(datasetpath, 'r') as f:
            lines = f.readlines()
            for line in lines:
                self.transactions.append(self.createTransaction(line))
        f.close()

    def createTransaction(self, line):
        """
        A method to create Transaction from dataset given

        :param line: represent a single line of database
        :type line: string
        :return : Transaction.
        :rtype: int
        """
        trans_list = line.strip().split(':')
        transactionUtility = int(trans_list[1])
        itemsString = trans_list[0].strip().split(self.sep)
        utilityString = trans_list[2].strip().split(self.sep)
        if (len(trans_list) == 4):
            pmuString = trans_list[3].strip().split(self.sep)
        items = []
        utilities = []
        pmus = []
        for idx, item in enumerate(itemsString):
            if (self.strToint).get(item) is None:
                self.strToint[item] = self.cnt
                self.intTostr[self.cnt] = item
                self.cnt += 1
            item_int = self.strToint.get(item)
            if item_int > self.maxItem:
                self.maxItem = item_int
            items.append(item_int)
            utilities.append(int(utilityString[idx]))
            if (len(trans_list) == 4):
                pmus.append(int(pmuString[idx]))
        return Transaction(items, utilities, transactionUtility, pmus)

    def getMaxItem(self):
        """
        A method to return name of the largest item
        """
        return self.maxItem

    def getTransactions(self):
        """
        A method to return transactions from database
        """
        return self.transactions


class TKSHUIM(utilityPatterns):
    """
    :Description:
       Top K Spatial High Utility ItemSet Mining (TKSHUIM) aims to discover Top-K Spatial High Utility Itemsets
       (TKSHUIs) in a spatioTemporal database

    :Reference:

       P. Pallikila et al., "Discovering Top-k Spatial High Utility Itemsets in Very Large Quantitative Spatiotemporal 
       databases," 2021 IEEE International Conference on Big Data (Big Data), Orlando, FL, USA, 2021, pp. 4925-4935, 
       doi: 10.1109/BigData52589.2021.9671912.

    :param  iFile: str :
                   Name of the Input file to mine complete set of High Utility Spatial patterns
    :param  oFile: str :
                   Name of the output file to store complete set of High Utility Spatial patterns
    :param minUtil: int :
                   Minimum utility threshold given by User
    :param maxMemory: int :
                   Maximum memory used by this program for running
    :param candidateCount: int :
                   Number of candidates to consider when calculating a high utility spatial pattern
    :param nFile: str :
                   Name of the input file to mine complete set of High Utility Spatial patterns
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.


    :Attributes:

        iFile : file
            Name of the input file to mine complete set of frequent patterns
        nFile : file
            Name of the Neighbours file that contain neighbours of items
        oFile : file
            Name of the output file to store complete set of frequent patterns
        memoryRSS : float
            To store the total amount of RSS memory consumed by the program
        startTime:float
            To record the start time of the mining process
        endTime:float
            To record the completion time of the mining process
        k : int
            The user given k value
        candidateCount: int
             Number of candidates 
        utilityBinArrayLU: list
             A map to hold the pmu values of the items in database
        utilityBinArraySU: list
            A map to hold the subtree utility values of the items is database
        oldNamesToNewNames: list
            A map to hold the subtree utility values of the items is database
        newNamesToOldNames: list
            A map to store the old name corresponding to new name
        Neighbours : map
            A dictionary to store the neighbours of a item
        maxMemory: float
            Maximum memory used by this program for running
        itemsToKeep: list
            keep only the promising items ie items having twu >= minUtil
        itemsToExplore: list
            keep items that subtreeUtility grater than minUtil

    :Methods:

        mine()
                Mining process will start from here
        getPatterns()
                Complete set of patterns will be retrieved with this function
        save(oFile)
                Complete set of patterns will be loaded in to a output file
        getPatternsAsDataFrame()
                Complete set of patterns will be loaded in to a dataframe
        getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
               Total amount of runtime taken by the mining process will be retrieved from this function
        calculateNeighbourIntersection(self, prefixLength)
               A method to return common Neighbours of items
        backtrackingEFIM(transactionsOfP, itemsToKeep, itemsToExplore, prefixLength)
               A method to mine the TKSHUIs Recursively
        useUtilityBinArraysToCalculateUpperBounds(transactionsPe, j, itemsToKeep, neighbourhoodList)
               A method to  calculate the sub-tree utility and local utility of all items that can extend itemSet P and e
        output(tempPosition, utility)
               A method ave a high-utility itemSet to file or memory depending on what the user chose
        is_equal(transaction1, transaction2)
               A method to Check if two transaction are identical
        intersection(lst1, lst2)
               A method that return the intersection of 2 list
        useUtilityBinArrayToCalculateSubtreeUtilityFirstTime(dataset)
              Scan the initial database to calculate the subtree utility of each items using a utility-bin array
        sortDatabase(self, transactions)
              A Method to sort transaction in the order of PMU
        sort_transaction(self, trans1, trans2)
              A Method to sort transaction in the order of PMU
        useUtilityBinArrayToCalculateLocalUtilityFirstTime(self, dataset)
             A method to scan the database using utility bin array to calculate the pmus                   

    **Executing the code on terminal:**
    -------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 TKSHUIM.py <inputFile> <outputFile> <Neighbours> <k> <sep>

      Example Usage:

      (.venv) $ python3 TKSHUIM.py sampleTDB.txt output.txt sampleN.txt 35

    .. note:: maxMemory will be considered as Maximum memory used by this program for running


    **Sample run of importing the code:**
    ----------------------------------------
    .. code-block:: python
        
            from PAMI.highUtilitySpatialPattern.topk import TKSHUIM as alg

            obj=alg.TKSHUIM("input.txt","Neighbours.txt",35)

            obj.mine()

            Patterns = obj.getPatterns()

            obj.save("output")

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ---------------
            The complete program was written by Pradeep Pallikila under the supervision of Professor Rage Uday Kiran.
    """
    candidateCount = 0
    utilityBinArrayLU = {}
    utilityBinArraySU = {}
    oldNamesToNewNames = {}
    newNamesToOldNames = {}
    strToint = {}
    intTostr = {}
    Neighbours = {}
    temp = [0] * 5000
    maxMemory = 0
    startTime = float()
    endTime = float()
    finalPatterns = {}
    iFile = " "
    oFile = " "
    nFile = " "
    sep = "\t"
    minUtil = 0
    memoryUSS = float()
    memoryRSS = float()
    heapList = []

    def __init__(self, iFile, nFile, k, sep="\t"):
        super().__init__(iFile, nFile, k, sep)

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Main function of the program.
        """
        self.mine()

    def mine(self):
        """
        Main function of the program.
        """
        self.startTime = time.time()
        self.finalPatterns = {}
        self.dataset = Dataset(self.iFile, self.sep)
        with open(self.nFile, 'r') as o:
            lines = o.readlines()
            for line in lines:
                line = line.split("\n")[0]
                line_split = line.split(self.sep)
                item = self.dataset.strToint.get(line_split[0])
                lst = []
                for i in range(1, len(line_split)):
                    lst.append(self.dataset.strToint.get(line_split[i]))
                self.Neighbours[item] = lst
        o.close()
        InitialMemory = psutil.virtual_memory()[3]
        self.useUtilityBinArrayToCalculateLocalUtilityFirstTime(self.dataset)
        itemsToKeep = []
        for key in self.utilityBinArrayLU.keys():
            if self.utilityBinArrayLU[key] >= self.minUtil:
                itemsToKeep.append(key)
        itemsToKeep = sorted(itemsToKeep, key=lambda x: self.utilityBinArrayLU[x])
        currentName = 1
        for idx, item in enumerate(itemsToKeep):
            self.oldNamesToNewNames[item] = currentName
            self.newNamesToOldNames[currentName] = item
            itemsToKeep[idx] = currentName
            currentName += 1
        for transaction in self.dataset.getTransactions():
            transaction.removeUnpromisingItems(self.oldNamesToNewNames)
        self.sortDatabase(self.dataset.getTransactions())
        emptyTransactionCount = 0
        for transaction in self.dataset.getTransactions():
            if len(transaction.getItems()) == 0:
                emptyTransactionCount += 1
        self.dataset.transactions = self.dataset.transactions[emptyTransactionCount:]
        self.useUtilityBinArrayToCalculateSubtreeUtilityFirstTime(self.dataset)
        self.heapList = []
        itemsToExplore = []
        for item in itemsToKeep:
            if self.utilityBinArraySU[item] >= self.minUtil:
                itemsToExplore.append(item)
        commonitems = []
        for i in range(self.dataset.maxItem):
            commonitems.append(i)
        self.backtrackingEFIM(self.dataset.getTransactions(), itemsToKeep, itemsToExplore, 0)
        finalMemory = psutil.virtual_memory()[3]
        memory = (finalMemory - InitialMemory) / 10000
        if memory > self.maxMemory:
            self.maxMemory = memory
        self.endTime = time.time()
        process = psutil.Process(os.getpid())
        self.memoryUSS = float()
        self.memoryRSS = float()
        self.memoryUSS = process.memory_full_info().uss
        self.memoryRSS = process.memory_info().rss
        for item in self.heapList:
            self.finalPatterns[item[1]] = item[0]
        print('TOP-K mining process is completed by TKSHUIM')

    def backtrackingEFIM(self, transactionsOfP, itemsToKeep, itemsToExplore, prefixLength):
        """
        A method to mine the TKSHUIs Recursively

        :param transactionsOfP: the list of transactions containing the current prefix P
        :type transactionsOfP: list
        :param itemsToKeep: the list of secondary items in the p-projected database
        :type itemsToKeep: list
        :param itemsToExplore: the list of primary items in the p-projected database
        :type itemsToExplore: list
        :param prefixLength: current prefixLength
        :type prefixLength: int
        """
        self.candidateCount += len(itemsToExplore)
        for idx, e in enumerate(itemsToExplore):
            initialMemory = psutil.virtual_memory()[3]
            transactionsPe = []
            utilityPe = 0
            if len(transactionsOfP) == 0:
                break 
            previousTransaction = transactionsOfP[0]
            consecutiveMergeCount = 0
            for transaction in transactionsOfP:
                items = transaction.getItems()
                if e in items:
                    positionE = items.index(e)
                    if transaction.getLastPosition() == positionE:
                        utilityPe += transaction.getUtilities()[positionE] + transaction.prefixUtility
                    else:
                        projectedTransaction = transaction.projectTransaction(positionE)
                        utilityPe += projectedTransaction.prefixUtility
                        if previousTransaction == transactionsOfP[0]:
                            previousTransaction = projectedTransaction
                        elif self.is_equal(projectedTransaction, previousTransaction):
                            if consecutiveMergeCount == 0:
                                items = previousTransaction.items[previousTransaction.offset:]
                                utilities = previousTransaction.utilities[previousTransaction.offset:]
                                itemsCount = len(items)
                                positionPrevious = 0
                                positionProjection = projectedTransaction.offset
                                while positionPrevious < itemsCount:
                                    utilities[positionPrevious] += projectedTransaction.utilities[positionProjection]
                                    positionPrevious += 1
                                    positionProjection += 1
                                previousTransaction.prefixUtility += projectedTransaction.prefixUtility
                                sumUtilities = previousTransaction.prefixUtility
                                previousTransaction = Transaction(items, utilities, previousTransaction.transactionUtility + projectedTransaction.transactionUtility)
                                previousTransaction.prefixUtility = sumUtilities
                            else:
                                positionPrevious = 0
                                positionProjected = projectedTransaction.offset
                                itemsCount = len(previousTransaction.items)
                                while positionPrevious < itemsCount:
                                    previousTransaction.utilities[positionPrevious] += projectedTransaction.utilities[
                                        positionProjected]
                                    positionPrevious += 1
                                    positionProjected += 1
                                previousTransaction.transactionUtility += projectedTransaction.transactionUtility
                                previousTransaction.prefixUtility += projectedTransaction.prefixUtility
                            consecutiveMergeCount += 1
                        else:
                            transactionsPe.append(previousTransaction)
                            previousTransaction = projectedTransaction
                            consecutiveMergeCount = 0
                    transaction.offset = positionE
            if previousTransaction != transactionsOfP[0]:
                transactionsPe.append(previousTransaction)
            self.temp[prefixLength] = self.newNamesToOldNames[e]
            if utilityPe >= self.minUtil:
                self.output(prefixLength, utilityPe)
            neighbourhoodList = self.calculateNeighbourIntersection(prefixLength)
            self.useUtilityBinArraysToCalculateUpperBounds(transactionsPe, idx, itemsToKeep, neighbourhoodList)
            newItemsToKeep = []
            newItemsToExplore = []
            for l in range(idx + 1, len(itemsToKeep)):
                itemK = itemsToKeep[l]
                if self.utilityBinArraySU[itemK] >= self.minUtil:
                    if itemK in neighbourhoodList:
                        newItemsToExplore.append(itemK)
                        newItemsToKeep.append(itemK)
                elif self.utilityBinArrayLU[itemK] >= self.minUtil:
                    if itemK in neighbourhoodList:
                        newItemsToKeep.append(itemK)
            self.backtrackingEFIM(transactionsPe, newItemsToKeep, newItemsToExplore, prefixLength + 1)
            finalMemory = psutil.virtual_memory()[3]
            memory = (finalMemory - initialMemory) / 10000
            if self.maxMemory < memory:
                self.maxMemory = memory

    def useUtilityBinArraysToCalculateUpperBounds(self, transactionsPe, j, itemsToKeep, neighbourhoodList):
        """
        A method to  calculate the sub-tree utility and local utility of all items that can extend itemSet P U {e}

        :param transactionsPe: transactions the projected database for P U {e}
        :type transactionsPe: list
        :param j:the position of j in the list of promising items
        :type j:int
        :param itemsToKeep :the list of promising items
        :type itemsToKeep: list
        :param neighbourhoodList: list of neighbourhood elements
        :type neighbourhoodList: list
        """
        for i in range(j + 1, len(itemsToKeep)):
            item = itemsToKeep[i]
            self.utilityBinArrayLU[item] = 0
            self.utilityBinArraySU[item] = 0
        for transaction in transactionsPe:
            length = len(transaction.getItems())
            i = length - 1
            while i >= transaction.offset:
                item = transaction.getItems()[i]
                if item in itemsToKeep:
                    remainingUtility = 0
                    if self.newNamesToOldNames[item] in self.Neighbours:
                        item_neighbours = self.Neighbours[self.newNamesToOldNames[item]]
                        for k in range(i, length):
                            transaction_item = transaction.getItems()[k]
                            if self.newNamesToOldNames[transaction_item] in item_neighbours and transaction_item in neighbourhoodList:
                                remainingUtility += transaction.getUtilities()[k]

                    remainingUtility += transaction.getUtilities()[i]
                    self.utilityBinArraySU[item] += remainingUtility + transaction.prefixUtility
                    self.utilityBinArrayLU[item] += transaction.transactionUtility + transaction.prefixUtility
                i -= 1

    def calculateNeighbourIntersection(self, prefixLength):
        """
        A method to find common Neighbours

        :param prefixLength: the prefix itemSet
        :type prefixLength:int
        """
        intersectionList = self.Neighbours.get(self.temp[0])
        for i in range(1, prefixLength+1):
            intersectionList = self.intersection(self.Neighbours[self.temp[i]], intersectionList)
        finalIntersectionList = []
        if intersectionList is None:
            return finalIntersectionList
        for item in intersectionList:
            if item in self.oldNamesToNewNames:
                finalIntersectionList.append(self.oldNamesToNewNames[item])
        return finalIntersectionList
    
    def output(self, tempPosition, utility):
        """
        A method save all high-utility itemSet to file or memory depending on what the user chose

        :param tempPosition: position of last item
        :type tempPosition : int
        :param utility: total utility of itemSet
        :type utility: int
        """
        s1 = str()
        for i in range(0, tempPosition+1):
            s1 += self.dataset.intTostr.get((self.temp[i]))
            if i != tempPosition:
                s1 += "\t"
        self.additemset(s1, utility)

    def is_equal(self, transaction1, transaction2):
        """
        A method to Check if two transaction are identical

        :param  transaction1: the first transaction.
        :type  transaction1: Transaction
        :param  transaction2:   the second transaction.
        :type  transaction2: Transaction
        :return : whether both are identical or not
        :rtype: bool
        """

        length1 = len(transaction1.items) - transaction1.offset
        length2 = len(transaction2.items) - transaction2.offset
        if length1 != length2:
            return False
        position1 = transaction1.offset
        position2 = transaction2.offset
        while position1 < len(transaction1.items):
            if transaction1.items[position1] != transaction2.items[position2]:
                return False
            position1 += 1
            position2 += 1
        return True
    
    def intersection(self, lst1, lst2):
        """
        A method that return the intersection of 2 list

        :param  lst1: items neighbour to item1
        :type lst1: list
        :param lst2: items neighbour to item2
        :type lst2: list
        :return :intersection of two lists
        :rtype : list
        """
        temp = set(lst2)
        lst3 = [value for value in lst1 if value in temp]
        return lst3

    def useUtilityBinArrayToCalculateSubtreeUtilityFirstTime(self, dataset):
        """
        Scan the initial database to calculate the subtree utility of each item using a utility-bin array

        :param dataset: the transaction database
        :type dataset: Dataset
        """
        for transaction in dataset.getTransactions():
            items = transaction.getItems()
            utilities = transaction.getUtilities()
            for idx, item in enumerate(items):
                if item not in self.utilityBinArraySU:
                    self.utilityBinArraySU[item] = 0
                if self.newNamesToOldNames[item] not in self.Neighbours:
                    self.utilityBinArraySU[item] += utilities[idx]
                    continue
                i = idx + 1
                sumSu = utilities[idx]
                while i < len(items):
                    if self.newNamesToOldNames[items[i]] in self.Neighbours[self.newNamesToOldNames[item]]:
                        sumSu += utilities[i]
                    i += 1
                self.utilityBinArraySU[item] += sumSu

    def sortDatabase(self, transactions):
        """
        A Method to sort transaction in the order of PMU

        :param transactions: transaction of items
        :type transactions: Transaction
        :return: sorted transaction
        :rtype: Transaction
        """
        cmp_items = cmp_to_key(self.sort_transaction)
        transactions.sort(key=cmp_items)

    def sort_transaction(self, trans1, trans2):
        """
        A Method to sort transaction in the order of PMU

        :param trans1: the first transaction.
        :type trans1: Transaction
        :param trans2:the second transaction.
        :type trans2: Transaction
        :return: sorted transaction.
        :rtype: int
        """
        trans1_items = trans1.getItems()
        trans2_items = trans2.getItems()
        pos1 = len(trans1_items) - 1
        pos2 = len(trans2_items) - 1
        if len(trans1_items) < len(trans2_items):
            while pos1 >= 0:
                sub = trans2_items[pos2] - trans1_items[pos1]
                if sub != 0:
                    return sub
                pos1 -= 1
                pos2 -= 1
            return -1
        elif len(trans1_items) > len(trans2_items):
            while pos2 >= 0:
                sub = trans2_items[pos2] - trans1_items[pos1]
                if sub != 0:
                    return sub
                pos1 -= 1
                pos2 -= 1
            return 1
        else:
            while pos2 >= 0:
                sub = trans2_items[pos2] - trans1_items[pos1]
                if sub != 0:
                    return sub
                pos1 -= 1
                pos2 -= 1
            return 0

    def useUtilityBinArrayToCalculateLocalUtilityFirstTime(self, dataset):
        """
        A method to scan the database using utility bin array to calculate the pmus

        :param dataset: the transaction database.
        :type dataset: database
        """
        utilityMatrix = defaultdict(lambda: defaultdict(int))
        for transaction in dataset.getTransactions():
            for idx, item in enumerate(transaction.getItems()):
                pmu = transaction.getUtilities()[idx]
                if item in self.Neighbours:
                    neighbors = self.Neighbours[item]
                    for idx, item in enumerate(transaction.getItems()):
                        if item in neighbors:
                            pmu += transaction.getUtilities()[idx]
                if item in self.utilityBinArrayLU:
                    # self.utilityBinArrayLU[item] += transaction.getPmus()[idx]
                    self.utilityBinArrayLU[item] += pmu
                else:
                    # self.utilityBinArrayLU[item] = transaction.getPmus()[idx]
                    self.utilityBinArrayLU[item] = pmu
                utilityMatrix[item][item] += transaction.getUtilities()[idx]
                if item in self.Neighbours:
                    neighbors = self.Neighbours[item]
                    utility = transaction.getUtilities()[idx]
                    for i, itemj in enumerate(transaction.getItems()):
                        if (itemj != item) and (itemj in neighbors):
                            utilityMatrix[item][itemj] += (utility + transaction.getUtilities()[i])

        for item in utilityMatrix.keys():
            for itemj in utilityMatrix[item].keys():
                if itemj >= item:
                    val = utilityMatrix[item][itemj]
                    if val != 0 and val > self.minUtil:
                        if itemj == item:
                            itemset = str(item)
                        else:
                            itemset = str(item) + str(itemj)
                        self.additemset(itemset, val)

    def additemset(self, itemset, utility):
        """
        adds the itemset to the priority queue

        :param itemset: the itemset to be added

        :type itemset: str

        :param utility: utility matrix for the itemset to be added

        :type utility: numpy.array
        """
        heapq.heappush(self.heapList, (utility, itemset))
        if len(self.heapList) > self.k:
            while len(self.heapList) > self.k:
                heapq.heappop(self.heapList)
                if len(self.heapList) == 0:
                    break
            self.minUtil = heapq.nsmallest(1, self.heapList)[0][0]

    def getPatternsAsDataFrame(self):
        """
        Storing final patterns in a dataframe

        :return: returning patterns in a dataframe
        :rtype: pd.DataFrame
        """
        dataFrame = {}
        data = []
        for a, b in self.finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataFrame = pd.DataFrame(data, columns=['Patterns', 'Utility'])

        return dataFrame
    
    def getPatterns(self):
        """
        Function to send the set of patterns after completion of the mining process

        :return: returning patterns
        :rtype: dict
        """
        return self.finalPatterns

    def save(self, outFile):
        """
        Complete set of patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self.finalPatterns.items():
            patternsAndSupport = x.strip() + ":" + str(y)
            writer.write("%s \n" % patternsAndSupport)

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self.memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """
        return self.memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self.endTime-self.startTime

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Top K Spatial  High Utility Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())

def main():
    inputFile = 'mushroom_utility_spmf.txt'
    neighborFile = 'mushroom_neighbourhood.txt' #Users can also specify this constraint between 0 to 1.
    k = 1000
    seperator = ' ' 
    obj = TKSHUIM(iFile=inputFile, nFile=neighborFile, k=k,  sep=seperator)    #initialize
    obj.mine()
    obj.mine()
    obj.printResults()
    print(obj.getPatterns())

if __name__ == '__main__':
    main()
    # _ap = str()
    # if len(sys.argv) == 5 or len(sys.argv) == 6:
    #     if len(sys.argv) == 6:
    #         _ap = TKSHUIM(sys.argv[1], sys.argv[3], int(sys.argv[4]), sys.argv[5])
    #     if len(sys.argv) == 5:
    #         _ap = TKSHUIM(sys.argv[1], sys.argv[3], int(sys.argv[4]))
    #     _ap.mine()
    #     _ap.mine()
    #     print("Top K Spatial  High Utility Patterns:", len(_ap.getPatterns()))
    #     _ap.save(sys.argv[2])
    #     print("Total Memory in USS:", _ap.getMemoryUSS())
    #     print("Total Memory in RSS",  _ap.getMemoryRSS())
    #     print("Total ExecutionTime in seconds:", _ap.getRuntime())
    # else:
    #     for i in [1000, 5000]:
    #         _ap = TKSHUIM('/Users/Likhitha/Downloads/mushroom_main_2000.txt',
    #                 '/Users/Likhitha/Downloads/mushroom_neighbors_2000.txt', i, ' ')
    #         _ap.mine()
    #         _ap.mine()
    #         print("Total number of Spatial High Utility Patterns:", len(_ap.getPatterns()))
    #         print("Total Memory in USS:", _ap.getMemoryUSS())
    #         print("Total Memory in RSS", _ap.getMemoryRSS())
    #         print("Total ExecutionTime in seconds:", _ap.getRuntime())
    #     print("Error! The number of input parameters do not match the total number of parameters provided")




# generateSpatioTemporalDatabase is a code used to convert the database into SpatioTemporal database.
#
#   **Importing this algorithm into a python program**
#   --------------------------------------------------------
#
#             from PAMI.extras.generateDatabase import generateSpatioTemporalDatabase as db
#
#             obj = db.generateSpatioTemporalDatabase(0, 100, 0, 100, 10, 10, 0.5, 0.9, 0.5, 0.9)
#
#             obj.save()
#
#             obj.createPoint(0,100,0,100) # values can be according to the size of data
#
#             obj.saveAsFile("outputFileName") # To create a file
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import random as rand
from typing import List, Dict, Tuple, Set, Union, Any, Generator
import pandas
import sys

class spatioTemporalDatabaseGenerator():
    """

    :Description:   generateSpatioTemporalDatabase is a code used to convert the database into SpatioTemporal database.

    :param  xmin: int :
            To give minimum value for x
    :param  xmax: int :
            To give maximum value for x
    :param  ymin: int :
            To give minimum value for y
    :param  ymax: int :
             To give maximum value for y
    :param maxTimeStamp: int :
            maximum Time Stamp for the database
    :param numberOfItems: int :
            number of items in the database
    :param itemChanceLow: int or float :
            least chance for item in the database
    :param itemChanceHigh: int or float :
            highest chance for item in the database
    :param timeStampChanceLow: int or float :
            lowest time stamp value
    :param timeStampChanceHigh: int or float:
            highest time stamp value

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.generateDatabase import generateSpatioTemporalDatabase as db

            obj = db.generateSpatioTemporalDatabase(0, 100, 0, 100, 10, 10, 0.5, 0.9, 0.5, 0.9)

            obj.save(oFile)

            obj.createPoint(0,100,0,100) # values can be according to the size of data

            obj.saveAsFile("outputFileName") # To create a file

    """

    coinFlip = [True, False]
    timestamp = list()
    items = list()
    alreadyAdded = set()
    outFileName=""

    def createPoint(self, xmin: int, xmax: int, ymin: int, ymax: int) -> Tuple[int, int]:
        x = rand.randint(xmin, xmax)
        y = rand.randint(ymin, ymax)
        coordinate = tuple([x, y])
        return coordinate

    def __init__(self,xmin: int,xmax: int,ymin: int,ymax: int,maxTimeStamp: int,numberOfItems: int, itemChanceLow: float,
                 itemChanceHigh: float, timeStampChanceLow: float,
                 timeStampChanceHigh: float) -> None:
        coinFlip = [True, False]
        timeStamp = 1
        self.timeStampList = list()
        self.itemList = list()

        while timeStamp != maxTimeStamp + 1:
            itemSet=list()
            for i in range(1, numberOfItems+1):
                #rand1=rand.rand(itemChanceLow,itemChanceHigh)
                #rand2 = rand.rand(timeStampChanceLow, timeStampChanceHigh)
                if rand.choices(coinFlip, weights=[itemChanceLow,itemChanceHigh], k=1)[0]:
                    coordinate=self.createPoint(xmin, xmax, ymin, ymax)
                    coordinate=tuple(coordinate)
                    if coordinate not in self.alreadyAdded:
                        coordinate=list(coordinate)
                        itemSet.append(coordinate)
                        coordinate=tuple(coordinate)
                        self.alreadyAdded.add(coordinate)
            if itemSet != []:
                self.timeStampList.append(
                    timeStamp)
                self.itemList.append(
                    itemSet)
            if rand.choices(coinFlip, weights=[itemChanceLow,itemChanceHigh], k=1)[0]:
                 timeStamp += 1
        self.outFileName = "temporal_" + str(maxTimeStamp // 1000) + \
                           "KI" + str(numberOfItems) + "C" + str(itemChanceLow) + "T" + str(timeStampChanceLow) + ".csv"




    def saveAsFile(self, outFileName="", sep="\t") -> None:
        if outFileName != "":
            self.outFileName = outFileName

        file = open(
            self.outFileName, "w")

        for i in range(len(self.timeStampList)):
            file.write(
                str(self.timeStampList[i]))
            for j in range(len(self.itemList[i])):
                file.write(
                    sep + str(self.itemList[i][j]))
            file.write('\n')

        file.close()


if __name__ == "__main__":
    xmin=0
    xmax=100
    ymin=0
    ymax=100
    maxTimeStamp = 10
    numberOfItems = 10
    itemChanceLow = 0.5
    itemChanceHigh = 0.9
    timeStampChanceLow = 0.5
    timeStampChanceHigh = 0.9
    obj = spatioTemporalDatabaseGenerator(sys.argv[1], sys.argv[2], sys.argv[3], sys.argv[4])
    obj.saveAsFile(sys.argv[5])



# createTDB in this code  we will create transactional Database.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.DF2DB import createTDB as ct
#
#             obj = ct.createTDB(idf, ">=", 16)
#
#             obj.save(oFile)
#




import sys

__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import pandas as pd


class createTDB:
    """
    :Description: This class will create Transactional database.

    :param df: It represents the dataframe
    :type df: list
    :param threshold : It is the threshold value of all item.
    :type threshold: int or float



    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.frequentPattern.basic import FPGrowth as fp

            obj = fp.createTDB(idf, ">=" )

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    """
    __startTime = float()
    __endTime = float()
    __memoryUSS = float()
    __memoryRSS = float()
    __Database = []
    __finalPatterns = {}

    def __init__(self, df, threshold):
        self._df = df
        self._threshold = int(threshold)
        self._items = []
        self._updatedItems = []

    def createTDB(self):
        """
        :Description:  To Create transactional database
        """
        i = self._df.columns.values.tolist()
        if 'sid' in i:
            self._items = self._df['sid'].tolist()
        for i in self._items:
            i = i.split()
            self._updatedItems.append([j for j in i if int(j) > self._threshold])

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x in self._updatedItems:
            s = str()
            for j in x:
                s = s + j + " "
            writer.write("%s \n" % s)


if __name__ == '__main__':
    a = createTDB(sys.argv[1], sys.argv[3])
    a.createTDB()
    a.save(sys.argv[2])


import pickle

class DfsCode:
    def __init__(self):
        """
        Initializes the DFSCode object with default values.
        """
        self.rightMost = -1
        self.size = 0
        self.rightMostPath = []  
        self.eeList = []

    def copy(self):
        """
        Creates a deep copy of the DFSCode object.
        """
        return pickle.loads(pickle.dumps(self))

    def notPreOfRm(self, v):
        """
            Checks if a given vertex is not the second-to-last element on the rightmost path.

            Args:
                v (int): The vertex to check.

            Returns:
                bool: True if the vertex is not the second-to-last element on the rightmost path, False otherwise.
            """
        if len(self.rightMostPath) <= 1:
            return True
        return v != self.rightMostPath[-2]

    def getAllVLabels(self):
        """
            Retrieves all vertex labels from the extended edge list.

            Returns:
                list: A list of vertex labels.
            """
        labels = []
        vertexMap = {}
        for ee in self.eeList:
            v1, v1Label = ee.getV1(), ee.getVLabel1()
            v2, v2Label = ee.getV2(), ee.getVLabel2()
            vertexMap[v1] = v1Label
            vertexMap[v2] = v2Label
        
        count = 0
        while count in vertexMap:
            labels.append(vertexMap[count])
            count += 1
        return labels

    def add(self, ee):
        """
            Adds an extended edge to the DFSCode object.

            Args:
                ee (ExtendedEdge): The extended edge to add.
            """
        if self.size == 0:
            self.rightMost = 1
            self.rightMostPath.extend([0, 1])
        else:
            v1, v2 = ee.getV1(), ee.getV2()
            if v1 < v2:
                self.rightMost = v2
                while self.rightMostPath and self.rightMostPath[-1] > v1:
                    self.rightMostPath.pop()
                self.rightMostPath.append(v2)

        self.eeList.append(ee)
        self.size += 1

    def getAt(self, index):
        """
            Retrieves the extended edge at the specified index.

            Args:
                index (int): The index of the extended edge to retrieve.

            Returns:
                ExtendedEdge: The extended edge at the specified index.
            """

        return self.eeList[index]

    def onRightMostPath(self, v):
        """
            Checks if a vertex is present on the rightmost path.

            Args:
                v (int): The vertex to check.

            Returns:
                bool: True if the vertex is present on the rightmost path, False otherwise.
            """
        return v in self.rightMostPath

    def containEdge(self, v1, v2):
        """
            Checks if an edge between vertices v1 and v2 is present.

            Args:
                v1 (int): The first vertex of the edge.
                v2 (int): The second vertex of the edge.

            Returns:
                bool: True if the edge is present, False otherwise.
            """
        for ee in self.eeList:
            if (ee.getV1() == v1 and ee.getV2() == v2) or (ee.getV1() == v2 and ee.getV2() == v1):
                return True
        return False

    def isEmpty(self):
        """
            Checks if the DFSCode object is empty.

            Returns:
                bool: True if the DFSCode object is empty, False otherwise."""
        return not self.eeList

    def getRightMost(self):
        """
            Retrieves the rightmost vertex.

            Returns:
                int: The index of the rightmost vertex.
            """
        return self.rightMost

    def getRightMostPath(self):
        """
            Retrieves the rightmost path.

            Returns:
                list: A list containing the vertices on the rightmost path.
            """
        return self.rightMostPath

    def getEeList(self):
        """
            Retrieves the extended edge list.

            Returns:
                list: A list of ExtendedEdge objects.
            """
        return self.eeList

    def __str__(self):
        """
            Returns a string representation of the DFSCode object.

            Returns:
                str: A string representation of the DFSCode object.
            """
        return "DfsCode: " + " ".join(str(ee) for ee in self.eeList)



# gSpan is a subgraph mining algorithm that uses DFS and DFS codes to mine subgraphs
#
# **Importing this algorithm into a python program**
#
#             from PAMI.subgraphMining.basic import gspan as alg
#
#             obj = alg.GSpan(iFile, minSupport)
#
#             obj.mine()
#
#             obj.run()
#
#             frequentGraphs = obj.getFrequentSubgraphs()
#
#             memUSS = obj.getMemoryUSS()
#
#             obj.save(oFile)
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#


__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
class FrequentSubgraph:
    def __init__(self, dfsCode, setOfGraphsIds, support):
        self.dfsCode = dfsCode
        self.setOfGraphsIds = setOfGraphsIds
        self.support = support

    def __eq__(self, other):
        if not isinstance(other, FrequentSubgraph):
            return NotImplemented
        return self.support == other.support

    def __lt__(self, other):
        if not isinstance(other, FrequentSubgraph):
            return NotImplemented
        return self.support < other.support

    def __gt__(self, other):
        if not isinstance(other, FrequentSubgraph):
            return NotImplemented
        return self.support > other.support




# Create WebHook for the Discord Channel: https://support.discord.com/hc/en-us/articles/228383668-Intro-to-Webhooks

# Copy the discord.Webhook.from_url

# Send messages

from discord import SyncWebhook


class discord():

    def __init__(self, url: str) -> None:
        self.url = url


    def send(self, message: str) -> None:
        try:
            webhook = SyncWebhook.from_url(self.url)
            webhook.send(message)
        except Exception as e:
            # Print any error messages to stdout
            print(e)


# SparseFormatDF in this code the dense dataframe is converting databases into different transactional, temporal, utility types.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.convert import sparseDF2DB as db
#
#             obj = db.sparseDF2DB(idf, ">=", 16)
#
#             obj.save(oFile)
#
#             obj.createTransactional("outputFileName") # To create transactional database
#
#             obj.createTemporal("outputFileName") # To create temporal database
#
#             obj.createUtility("outputFileName") # To create utility database
#
#             obj.getFileName("outputFileName") # To get file name of the database
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import pandas as pd
import sys

class sparseDF2DB:
    """
    :Description:  This class create Data Base from DataFrame.

    :Attributes:

        :param inputDF: dataframe :
            It is dense DataFrame
        :param condition: str :
            It is condition to judge the value in dataframe
        :param thresholdValue: int or float :
            User defined value.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.DF2DB import SparseFormatDF as db

            obj = db.SparseFormatDF(iDdf, ">=", 16)

            obj.save(oFile)

            obj.createTransactional("outputFileName") # To create transactional database

            obj.createTemporal("outputFileName") # To create temporal database

            obj.createUtility("outputFileName") # To create utility database

            obj.getFileName("outputFileName") # To get file name of the database
    """


    def __init__(self, inputDF, condition: str, thresholdValue: float) -> None:
        self.inputDF = inputDF
        self.condition = condition
        self.thresholdValue = thresholdValue
        self.outputFile = ''
        if self.condition == '>':
            self.df = self.inputDF.query(f'value > {self.thresholdValue}')
        elif self.condition == '>=':
            self.df = self.inputDF.query(f'value >= {self.thresholdValue}')
        elif self.condition == '<=':
            self.df = self.inputDF.query(f'value <= {self.thresholdValue}')
        elif self.condition == '<':
            self.df = self.inputDF.query(f'value < {self.thresholdValue}')
        else:
            print('Condition error')
        self.df = self.df.drop(columns='value')
        self.df = self.df.groupby('tid')['item'].apply(list)

    def createTransactional(self, outputFile: str) -> None:
        """
        Create transactional data base
        :param outputFile: Write transactional data base into outputFile
        :type outputFile: str
        :return: None
        """
        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for line in self.df:
                f.write(f'{line[0]}')
                for item in line[1:]:
                    f.write(f',{item}')
                f.write('\n')

    def createTemporal(self, outputFile: str) -> None:
        """
        Create temporal data base
        :param outputFile: Write temporal data base into outputFile
        :type outputFile: str
        :return: None
        """

        self.outputFile = outputFile
        with open(self.outputFile, 'w') as f:
            for tid in self.df.index:
                f.write(f'{tid}')
                for item in self.df[tid]:
                    f.write(f',{item}')
                f.write('\n')

    def createUtility(self, outputFile: str) -> None:
        """
        Create the utility database.
        :param outputFile: Write utility database into outputFile
        :type outputFile: str
        :return: None
        """

        self.outputFile = outputFile
        items = self.inputDF.groupby(level=0)['item'].apply(list)
        values = self.inputDF.groupby(level=0)['value'].apply(list)
        sums = self.inputDF.groupby(level=0)['value'].sum()
        index = list(items.index)
        with open(self.outputFile, 'w') as f:
            for tid in index:
                f.write(f'{items[tid][0]}')
                for item in items[tid][1:]:
                    f.write(f'\t{item}')
                f.write(f':{sums[tid]}:')
                f.write(f'{values[tid][0]}')
                for value in values[tid][1:]:
                    f.write(f'\t{value}')
                f.write('\n')

    def getFileName(self) -> str:

        return self.outputFile

if __name__ == '__main__':

    obj = sparseDF2DB(sys.argv[1], sys.argv[2],sys.argv[3])
    obj.getFileName(sys.argv[4])




# EFIM is one of the fastest algorithm to mine High Utility ItemSets from transactional databases.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             from PAMI.highUtilitySpatialPattern.basic import efimParallel as alg
#
#             obj=alg.efimParallel("input.txt","Neighbours.txt",35)
#
#             obj.mine()
#
#             Patterns = obj.getPatterns()
#
#             print("Total number of Spatial High-Utility Patterns:", len(Patterns))
#
#             obj.save("output")
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

import os
import mmap
import time
import psutil
from joblib import Parallel, delayed
from deprecated import deprecated

__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

from PAMI.highUtilityPattern.basic import abstract as _ab

class efimParallel(_ab._utilityPatterns):
    """
    :Description:  EFIM is one of the fastest algorithm to mine High Utility ItemSets from transactional databases.
    
    :Reference:    Zida, S., Fournier-Viger, P., Lin, J.CW. et al. EFIM: a fast and memory efficient algorithm for
                   high-utility itemset mining. Knowl Inf Syst 51, 595–625 (2017). https://doi.org/10.1007/s10115-016-0986-0

    :param  iFile: str :
                   Name of the Input file to mine complete set of High Utility patterns
    :param  oFile: str :
                   Name of the output file to store complete set of High Utility patterns
    :param minUtil: int :
                   The user given minUtil value.
    :param candidateCount: int
                   Number of candidates specified by user
    :param maxMemory: int
                   Maximum memory used by this program for running
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        inputFile (str):
            The input file path.
        minUtil (int):
            The minimum utility threshold.
        sep (str):
            The separator used in the input file.
        threads (int):
            The number of threads to use.
        Patterns (dict):
            A dictionary containing the discovered patterns.
        rename (dict):
            A dictionary containing the mapping between the item IDs and their names.
        runtime (float):
            The runtime of the algorithm in seconds.
        memoryRSS (int):
            The Resident Set Size (RSS) memory usage of the algorithm in bytes.
        memoryUSS (int):
            The Unique Set Size (USS) memory usage of the algorithm in bytes.

    :Methods:

        read_file():
            Read the input file and return the filtered transactions, primary items, and secondary items.
        binarySearch(arr, item):
            Perform a binary search on the given array to find the given item.
        project(beta, file_data, secondary):
            Project the given beta itemset on the given database.
        search(collections):
            Search for high utility itemsets in the given collections.
        mine():
            Start the EFIM algorithm.
        savePatterns(outputFile):
            Save the patterns discovered by the algorithm to an output file.
        getPatterns():
            Get the patterns discovered by the algorithm.
        getRuntime():
            Get the runtime of the algorithm.
        getMemoryRSS():
            Get the Resident Set Size (RSS) memory usage of the algorithm.
        getMemoryUSS():
            Get the Unique Set Size (USS) memory usage of the algorithm.
        printResults():
            Print the results of the algorithm.

    **Executing the code on terminal:**
    ------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 efimParallel.py <inputFile> <outputFile> <minUtil> <sep>

      Example Usage:

      (.venv) $ python3 efimParallel sampleTDB.txt output.txt 35

    .. note:: maxMemory will be considered as Maximum memory used by this program for running


    **Importing this algorithm into a python program**
    -------------------------------------------------------
    .. code-block:: python

            from PAMI.highUtilityPattern.basic import efimParallel as alg

            obj = alg.efimParallel("input.txt",35)

            obj.mine()

            Patterns = obj.getPatterns()

            print("Total number of high utility Patterns:", len(Patterns))

            obj.save("output")

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ----------------
            The complete program was written by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.
    """

    def __init__(self, iFile, minUtil, sep="\t", threads=1):
        super().__init__(iFile, minUtil, sep)
        self.inputFile = iFile
        self.minUtil = minUtil
        self.sep = sep
        self.Patterns = {}
        self.rename = {}
        self.threads = threads

    # Read input file
    def _read_file(self):
        """
        Read the input file and return the filtered transactions, primary items, and secondary items.

        :return:

            filtered_transactions (dict): A dictionary containing the filtered transactions.
            primary (set): A set containing the primary items.
            secondary (set): A set containing the secondary items.
        """


        file_data = []
        twu = {}

        with open(self.inputFile, 'r') as f:
            fd = mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ)

            for line in iter(fd.readline, b""):
                line = line.decode('utf-8').strip().split(":")
                
                # Parse and process the line
                line = [x.split(self.sep) for x in line]
                weight = int(line[1][0])

                # Update file data with the parsed items
                file_data.append([line[0], [int(x) for x in line[2]]])

                for k in line[0]:
                    if k not in twu:
                        twu[k] = weight
                    else:
                        twu[k] += weight

        # Filter TWU dictionary based on minUtil (minimum utility threshold)
        twu = {k: v for k, v in twu.items() if v >= self.minUtil}

        # Sort TWU items by utility
        twu = {k: v for k, v in sorted(twu.items(), key=lambda item: item[1], reverse=True)}

        strToInt = {}
        t = len(twu)
        for k in twu.keys():
            strToInt[k] = t
            self.rename[t] = k
            t -= 1

        secondary = set(self.rename.keys())

        # Filter and sort transactions
        subtree = {}
        filtered_transactions = {}
        for col in file_data:
            zipped = zip(col[0], col[1])
            transaction = [(strToInt[x], y) for x, y in zipped if x in strToInt]
            transaction = sorted(transaction, key=lambda x: x[0])
            if len(transaction) > 0:
                val = [x[1] for x in transaction]
                key = [x[0] for x in transaction]
                
                fs = frozenset(key)

                if fs not in filtered_transactions:
                    filtered_transactions[fs] = [key, val, 0]
                else:
                    filtered_transactions[fs][1] = [x + y for x, y in zip(filtered_transactions[fs][1], val)]

                subUtil = sum([x[1] for x in transaction])
                temp = 0

                for i in range(len(transaction)):
                    item = key[i]
                    if item not in subtree:
                        subtree[item] = subUtil - temp
                    else:
                        subtree[item] += subUtil - temp
                    temp += val[i]

        primary = [key for key in subtree.keys() if subtree[key] >= self.minUtil]

        return filtered_transactions, primary, secondary
    

    def _binarySearch(self, arr, item):
        """
        Do a binary search on the given array to find the given item.

        :param arr: The array to search in.

        :type arr: list

        :param item: The item to search for

        :type item: int
        """

        low = 0
        high = len(arr) - 1
        mid = 0

        while low <= high:
            mid = (high + low) // 2
            if arr[mid] < item:
                low = mid + 1
            elif arr[mid] > item:
                high = mid - 1
            else:
                return mid

        return -1

    def _project(self, beta, file_data, secondary):
        """
        Project the given beta itemset on the given database.

        :param beta: The beta itemset to be projected

        :type beta: list

        :param file_data: The database to project on.

        :type file_data: list

        :param secondary: The set of secondary items

        :type secondary: set

        :return:

            projected_db (dict):
                The projected database.
            local_utils (dict):
                The local utilities of the projected database.
            subtree_utils (dict):
                The subtree utilities of the projected database.
            utility (int):
                The utility of the projected database.
        """


        projected_db = {}
        local_utils = {}
        subtree_utils = {}
        utility = 0     

        added = set()

        item = beta[-1]

        temp = [v for k, v in file_data.items() if item in k]
        start = time.time()

        for v in temp:
            index = self._binarySearch(v[0], item)

            curr = v[1][index] + v[2]
            utility += curr

            newKey = []
            newVal = []

            for i in range(index+1, len(v[0])):
                if v[0][i] in secondary:
                    newKey.append(v[0][i])
                    newVal.append(v[1][i])

            if len(newKey) == 0:
                continue

            s = sum(newVal)
            temp = 0

            for i in range(len(newKey)):
                if newKey[i] in added:
                    local_utils[newKey[i]] += s + curr
                    subtree_utils[newKey[i]] += s + curr - temp
                else:
                    local_utils[newKey[i]] = s + curr
                    subtree_utils[newKey[i]] = s + curr - temp
                    added.add(newKey[i])
                
                temp += newVal[i]
            
            fs = frozenset(newKey)

            if fs not in projected_db:
                projected_db[fs] = [newKey, newVal, curr]
            else:
                projected_db[fs][1] = [x + y for x, y in zip(projected_db[fs][1], newVal)]
                projected_db[fs][2] += curr

        nprimary = [key for key in subtree_utils.keys() if subtree_utils[key] >= self.minUtil]
        nsecondary = set([key for key in local_utils.keys() if local_utils[key] >= self.minUtil])

        return beta, projected_db, nsecondary, nprimary, utility
    

    def _search(self, collections):

        """
        Search for frequent patterns in the given collections.

        :param collections: list of collections to search in

        :type collections: list
        """

        if (self.threads > 1):
            with Parallel(n_jobs=self.threads) as parallel:
                while len(collections) > 0:
                    new_collections = []

                    # print("Num of tasks:", sum(len(collections[i][2]) for i in range(len(collections))))
                    results = parallel(delayed(self._project)(collections[i][0] + [collections[i][2][j]], collections[i][1], collections[i][3]) for i in range(len(collections)) for j in range(len(collections[i][2])))

                    for i in range(len(results)):
                        beta, projected_db, secondary, primary, utility = results[i]
                        if utility >= self.minUtil:
                            pattern = "\t".join([self.rename[x] for x in beta])
                            # self.Patterns[tuple(beta)] = utility
                            self.Patterns[pattern] = utility
                        if len(primary) > 0:
                            new_collections.append([beta, projected_db, primary, secondary])
                    
                    collections = new_collections

        else:
            while len(collections) > 0:
                new_collections = []
                for i in range(len(collections)):
                    for j in range(len(collections[i][2])):
                        beta, projected_db, secondary, primary, utility = self._project(collections[i][0] + [collections[i][2][j]], collections[i][1], collections[i][3])
                        if utility >= self.minUtil:
                            pattern = "\t".join([self.rename[x] for x in beta])
                            # self.Patterns[tuple(beta)] = utility
                            self.Patterns[pattern] = utility
                        if len(primary) > 0:
                            new_collections.append([beta, projected_db, primary, secondary])

                collections = new_collections

    @deprecated("It is recommended to use 'mine()' instead of 'mine()' for mining process. Starting from January 2025, 'mine()' will be completely terminated.")
    def startMine(self):
        """
        Start the EFIM algorithm.
        """

        self.mine()

    def mine(self):
        """
        Start the EFIM algorithm.
        """

        ps = psutil.Process(os.getpid())

        self.start = time.time()

        fileData, primary, secondary = self._read_file()

        collection = [[[], fileData, primary, secondary]]

        self._search(collection)

        self.memoryRSS = ps.memory_info().rss
        self.memoryUSS = ps.memory_full_info().uss

        end = time.time()
        self.runtime = end - self.start

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file
        :param outFile: name of the output file
        :type outFile: csv file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self._finalPatterns.items():
            patternsAndSupport = x.strip() + ":" + str(y)
            writer.write("%s \n" % patternsAndSupport)
    
    def getPatternsAsDataFrame(self):
        """
        Storing final patterns in a dataframe
        :return: returning patterns in a dataframe
        :rtype: pd.DataFrame
        """
        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a.replace('\t', ' '), b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Utility'])

        return dataFrame

    def getPatterns(self):
        """
        Get the patterns discovered by the algorithm.

        :return:
            dict: A dictionary containing the discovered patterns.
        """
        return self.Patterns

    def getRuntime(self):
        """
        Get the runtime of the algorithm.

        :return:
            float: The runtime in seconds.
        """
        return self.runtime

    def getMemoryRSS(self):
        """
        Get the Resident Set Size (RSS) memory usage of the algorithm.

        :return:
            int: The RSS memory usage in bytes.
        """
        return self.memoryRSS

    def getMemoryUSS(self):
        """
        Get the Unique Set Size (USS) memory usage of the algorithm.

        :return:
            int: The USS memory usage in bytes.
        """
        return self.memoryUSS


    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of High Utility Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in seconds:", self.getRuntime())


if __name__ == "__main__":

    # inputFile = 'EFIM/accidents_utility_spmf.txt'
    # minUtil = 28000000

    # # inputFile = 'EFIM/chainstore.txt'
    # # minUtil = 2500000

    # # inputFile = 'test.txt'
    # # minUtil = 5

    # # inputFile = "EFIM/BMS_utility_spmf.txt"
    # # minUtil = 2025000

    # # inputFile = "EFIM/Utility_pumsb.csv"
    # # minUtil = 7500000

    # sep = " "
    # f = efimParallel(inputFile, minUtil, sep, 1)
    # f.mine()
    # f.mine()
    # print("# of patterns: " + str(len(f.getPatterns())))
    # print("Time taken: " + str(f.getRuntime()))
    # f.savePatterns("mine.txt")


    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:    #includes separator
            _ap = efimParallel(_ab._sys.argv[1], int(_ab._sys.argv[3]), _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:    #takes "\t" as a separator
            _ap = efimParallel(_ab._sys.argv[1], int(_ab._sys.argv[3]))
        _ap.mine()
        _ap.mine()
        print("Total number of High Utility Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS",  _ap.getMemoryRSS())
        print("Total ExecutionTime in seconds:", _ap.getRuntime())
    else:
        _ap = efimParallel('/Users/likhitha/Downloads/Utility_T10I4D100K.csv', 50000, '\t')
        _ap.mine()
        _ap.mine()
        print("Total number of High Utility Patterns:", len(_ap.getPatterns()))
        _ap.save('/Users/likhitha/Downloads/UPGrowth_output.txt')
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")




import random


class generateUncertainTemporal:
    __transactionSize: int
    __numOfItems: int
    __avgTransactionLength: int
    __significant: int
    __transactions: list[list[int]]
    __probabilitis: list[list[float]]

    def __init__(self, transactionSize: int, numOfItems: int, avgTransactionLength: int, significant=2) -> None:
        self.__transactionSize = transactionSize
        self.__numOfItems = numOfItems
        self.__avgTransactionLength = avgTransactionLength
        self.__significant = significant

        self.__transactions = list()
        self.__probabilitis = list()

    def generate(self) -> None:
        for tid in range(self.__transactionSize):
            length = random.randint(1, self.__avgTransactionLength * 2)
            transaction = [random.randint(1, self.__numOfItems)
                           for _ in range(length)]
            self.__transactions.append(transaction)
            probability = [round(random.uniform(0, 1), self.__significant)
                           for _ in range(length)]
            self.__probabilitis.append(probability)

    def save(self, outputFile: str, sep="\t") -> None:
        with open(outputFile, 'w') as f:
            for tid, (transaction, probability) in enumerate(zip(self.__transactions, self.__probabilitis)):
                f.write(
                    f"{tid+1}{sep}{sep.join(map(str, transaction))}:{round(sum(probability), self.__significant)}:{sep.join(map(str, probability))}\n")


if __name__ == "__main__":
    obj = generateUncertainTemporal(10, 10, 5)
    obj.generate()
    obj.save("uncertainTemporal_test.csv")
